import torch
import torch.nn as nn
import numpy as np
from deepchem.models.torch_models import layers
from deepchem.models.torch_models.torch_model import TorchModel
from deepchem.models.losses import L2Loss
from typing import Any


class MAT(nn.Module):
    """An internal TorchModel class.

    In this class, we define the various layers and establish a sequential model for the Molecular Attention Transformer.
    We also define the forward call of this model in the forward function.

    References
    ----------
    .. [1] Lukasz Maziarka et al. "Molecule Attention Transformer" Graph Representation Learning workshop and Machine Learning and the Physical Sciences workshop at NeurIPS 2019. 2020. https://arxiv.org/abs/2002.08264

    Examples
    --------
    >>> import deepchem as dc
    >>> import pandas as pd
    >>> import numpy as np
    >>> smiles = ['CC', 'CCC',  'CCCC', 'CCCCC', 'CCCCCCC']
    >>> vals = [1.35, 6.72, 5.67, 1.23, 1.76]
    >>> df = pd.DataFrame(list(zip(smiles, vals)), columns = ['smiles', 'y'])
    >>> loader = dc.data.CSVLoader(tasks=['y'], feature_field='smiles', featurizer=dc.feat.MATFeaturizer())
    >>> df.to_csv('test.csv')
    >>> dataset = loader.create_dataset('test.csv')
    >>> model = dc.models.torch_models.MAT()
    >>> # To simulate input data, we will generate matrices for a single molecule.
    >>> vals = dataset.X[0]
    >>> node = vals.node_features
    >>> adj = vals.adjacency_matrix
    >>> dist = vals.distance_matrix
    >>> # We will now utilize a helper function defined in MATModel to get our matrices ready, and convert them into a batch consisting of a single molecule.
    >>> helper = dc.models.torch_models.MATModel()
    >>> node_features = helper.pad_sequence(torch.tensor(node).unsqueeze(0).float())
    >>> adjacency = helper.pad_sequence(torch.tensor(adj).unsqueeze(0).float())
    >>> distance = helper.pad_sequence(torch.tensor(dist).unsqueeze(0).float())
    >>> inputs = [node_features, adjacency, distance]
    >>> inputs = [x.astype(np.float32) if x.dtype == np.float64 else x for x in inputs]
    >>> # Get the forward call of the model for this batch.
    >>> output = model(inputs)
    """

    def __init__(self,
                 dist_kernel: str = 'softmax',
                 n_encoders=8,
                 lambda_attention: float = 0.33,
                 lambda_distance: float = 0.33,
                 h: int = 16,
                 sa_hsize: int = 1024,
                 sa_dropout_p: float = 0.0,
                 output_bias: bool = True,
                 d_input: int = 1024,
                 d_hidden: int = 1024,
                 d_output: int = 1024,
                 activation: str = 'leakyrelu',
                 n_layers: int = 1,
                 ff_dropout_p: float = 0.0,
                 encoder_hsize: int = 1024,
                 encoder_dropout_p: float = 0.0,
                 embed_input_hsize: int = 36,
                 embed_dropout_p: float = 0.0,
                 gen_aggregation_type: str = 'mean',
                 gen_dropout_p: float = 0.0,
                 gen_n_layers: int = 1,
                 gen_attn_hidden: int = 128,
                 gen_attn_out: int = 4,
                 gen_d_output: int = 1,
                 **kwargs):
        """
        Initialization for the internal MAT class.

        Parameters
        ----------
        dist_kernel: str
            Kernel activation to be used. Can be either 'softmax' for softmax or 'exp' for exponential, for the self-attention layer.
        n_encoders: int
            Number of encoder layers in the encoder block.
        lambda_attention: float
            Constant to be multiplied with the attention matrix in the self-attention layer.
        lambda_distance: float
            Constant to be multiplied with the distance matrix in the self-attention layer.
        h: int
            Number of attention heads for the self-attention layer.
        sa_hsize: int
            Size of dense layer in the self-attention layer.
        sa_dropout_p: float
            Dropout probability for the self-attention layer.
        output_bias: bool
            If True, dense layers will use bias vectors in the self-attention layer.
        d_input: int
            Size of input layer in the feed-forward layer.
        d_hidden: int
            Size of hidden layer in the feed-forward layer. Will also be used as d_output for the MATEmbedding layer.
        d_output: int
            Size of output layer in the feed-forward layer.
        activation: str
            Activation function to be used in the feed-forward layer.
            Can choose between 'relu' for ReLU, 'leakyrelu' for LeakyReLU, 'prelu' for PReLU,
            'tanh' for TanH, 'selu' for SELU, 'elu' for ELU and 'linear' for linear activation.
        n_layers: int
            Number of layers in the feed-forward layer.
        ff_dropout_p: float
            Dropout probability in the feeed-forward layer.
        encoder_hsize: int
            Size of Dense layer for the encoder itself.
        encoder_dropout_p: float
            Dropout probability for connections in the encoder layer.
        embed_input_hsize: int
            Size of input layer for the MATEmbedding layer.
        embed_dropout_p: float
            Dropout probability for the MATEmbedding layer.
        gen_aggregation_type: str
            Type of aggregation to be used. Can be 'grover', 'mean' or 'contextual'.
        gen_dropout_p: float
            Dropout probability for the MATGenerator layer.
        gen_n_layers: int
            Number of layers in MATGenerator.
        gen_attn_hidden: int
            Size of hidden attention layer in the MATGenerator layer.
        gen_attn_out: int
            Size of output attention layer in the MATGenerator layer.
        gen_d_output: int
            Size of output layer in the MATGenerator layer.
        """

        super(MAT, self).__init__()

        self.embedding = layers.MATEmbedding(d_input=embed_input_hsize,
                                             d_output=d_hidden,
                                             dropout_p=embed_dropout_p)

        self.encoder = nn.ModuleList([
            layers.MATEncoderLayer(dist_kernel=dist_kernel,
                                   lambda_attention=lambda_attention,
                                   lambda_distance=lambda_distance,
                                   h=h,
                                   sa_hsize=sa_hsize,
                                   sa_dropout_p=sa_dropout_p,
                                   output_bias=output_bias,
                                   d_input=d_input,
                                   d_hidden=d_hidden,
                                   d_output=d_output,
                                   activation=activation,
                                   n_layers=n_layers,
                                   ff_dropout_p=ff_dropout_p,
                                   encoder_hsize=encoder_hsize,
                                   encoder_dropout_p=encoder_dropout_p)
            for _ in range(n_encoders)
        ])

        self.generator = layers.MATGenerator(
            hsize=d_input,
            aggregation_type=gen_aggregation_type,
            d_output=gen_d_output,
            n_layers=gen_n_layers,
            dropout_p=gen_dropout_p,
            attn_hidden=gen_attn_hidden,
            attn_out=gen_attn_out)

    def forward(self, data: np.ndarray, **kwargs):
        node_features = torch.tensor(data[0]).float()
        adjacency_matrix = torch.tensor(data[1]).float()
        distance_matrix = torch.tensor(data[2]).float()

        mask = torch.sum(torch.abs(node_features), dim=-1) != 0
        output = self.embedding(node_features)

        for layer in self.encoder:
            output = layer(output, mask, adjacency_matrix, distance_matrix)
        output = self.generator(output, mask)
        return output


class MATModel(TorchModel):
    """Molecular Attention Transformer.

    This class implements the Molecular Attention Transformer [1]_.
    The MATFeaturizer (deepchem.feat.MATFeaturizer) is intended to work with this class.
    The model takes a batch of MATEncodings (from MATFeaturizer) as input, and returns an array of size Nx1, where N is the number of molecules in the batch.
    Each molecule is broken down into its Node Features matrix, adjacency matrix and distance matrix.
    A mask tensor is calculated for the batch. All of this goes as input to the MATEmbedding, MATEncoder and MATGenerator layers, which are defined in deepchem.models.torch_models.layers.py

    Currently, MATModel is intended to be a regression model for the freesolv dataset.

    References
    ----------
    .. [1] Lukasz Maziarka et al. "Molecule Attention Transformer" Graph Representation Learning workshop and Machine Learning and the Physical Sciences workshop at NeurIPS 2019. 2020. https://arxiv.org/abs/2002.08264

    Examples
    --------
    >>> import deepchem as dc
    >>> import pandas as pd
    >>> smiles = ['CC', 'CCC',  'CCCC', 'CCCCC', 'CCCCCCC']
    >>> vals = [1.35, 6.72, 5.67, 1.23, 1.76]
    >>> df = pd.DataFrame(list(zip(smiles, vals)), columns = ['smiles', 'y'])
    >>> loader = dc.data.CSVLoader(tasks=['y'], feature_field='smiles', featurizer=dc.feat.MATFeaturizer())
    >>> df.to_csv('test.csv')
    >>> dataset = loader.create_dataset('test.csv')
    >>> model = dc.models.torch_models.MATModel(batch_size = 2)
    >>> out = model.fit(dataset, nb_epoch = 1)
    """

    def __init__(self,
                 dist_kernel: str = 'softmax',
                 n_encoders=8,
                 lambda_attention: float = 0.33,
                 lambda_distance: float = 0.33,
                 h: int = 16,
                 sa_hsize: int = 1024,
                 sa_dropout_p: float = 0.0,
                 output_bias: bool = True,
                 d_input: int = 1024,
                 d_hidden: int = 1024,
                 d_output: int = 1024,
                 activation: str = 'leakyrelu',
                 n_layers: int = 1,
                 ff_dropout_p: float = 0.0,
                 encoder_hsize: int = 1024,
                 encoder_dropout_p: float = 0.0,
                 embed_input_hsize: int = 36,
                 embed_dropout_p: float = 0.0,
                 gen_aggregation_type: str = 'mean',
                 gen_dropout_p: float = 0.0,
                 gen_n_layers: int = 1,
                 gen_attn_hidden: int = 128,
                 gen_attn_out: int = 4,
                 gen_d_output: int = 1,
                 **kwargs):
        """The wrapper class for the Molecular Attention Transformer.

        Since we are using a custom data class as input (MATEncoding), we have overriden the default_generator function from DiskDataset and customized it to work with a batch of MATEncoding classes.

        Parameters
        ----------
        dist_kernel: str
            Kernel activation to be used. Can be either 'softmax' for softmax or 'exp' for exponential, for the self-attention layer.
        n_encoders: int
            Number of encoder layers in the encoder block.
        lambda_attention: float
            Constant to be multiplied with the attention matrix in the self-attention layer.
        lambda_distance: float
            Constant to be multiplied with the distance matrix in the self-attention layer.
        h: int
            Number of attention heads for the self-attention layer.
        sa_hsize: int
            Size of dense layer in the self-attention layer.
        sa_dropout_p: float
            Dropout probability for the self-attention layer.
        output_bias: bool
            If True, dense layers will use bias vectors in the self-attention layer.
        d_input: int
            Size of input layer in the feed-forward layer.
        d_hidden: int
            Size of hidden layer in the feed-forward layer. Will also be used as d_output for the MATEmbedding layer.
        d_output: int
            Size of output layer in the feed-forward layer.
        activation: str
            Activation function to be used in the feed-forward layer.
            Can choose between 'relu' for ReLU, 'leakyrelu' for LeakyReLU, 'prelu' for PReLU,
            'tanh' for TanH, 'selu' for SELU, 'elu' for ELU and 'linear' for linear activation.
        n_layers: int
            Number of layers in the feed-forward layer.
        ff_dropout_p: float
            Dropout probability in the feeed-forward layer.
        encoder_hsize: int
            Size of Dense layer for the encoder itself.
        encoder_dropout_p: float
            Dropout probability for connections in the encoder layer.
        embed_input_hsize: int
            Size of input layer for the MATEmbedding layer.
        embed_dropout_p: float
            Dropout probability for the MATEmbedding layer.
        gen_aggregation_type: str
            Type of aggregation to be used. Can be 'grover', 'mean' or 'contextual'.
        gen_dropout_p: float
            Dropout probability for the MATGenerator layer.
        gen_n_layers: int
            Number of layers in MATGenerator.
        gen_attn_hidden: int
            Size of hidden attention layer in the MATGenerator layer.
        gen_attn_out: int
            Size of output attention layer in the MATGenerator layer.
        gen_d_output: int
            Size of output layer in the MATGenerator layer.
        """
        model = MAT(dist_kernel=dist_kernel,
                    n_encoders=n_encoders,
                    lambda_attention=lambda_attention,
                    lambda_distance=lambda_distance,
                    h=h,
                    sa_hsize=sa_hsize,
                    sa_dropout_p=sa_dropout_p,
                    output_bias=output_bias,
                    d_input=d_input,
                    d_hidden=d_hidden,
                    d_output=d_output,
                    activation=activation,
                    n_layers=n_layers,
                    ff_dropout_p=ff_dropout_p,
                    encoder_hsize=encoder_hsize,
                    encoder_dropout_p=encoder_dropout_p,
                    embed_input_hsize=embed_input_hsize,
                    embed_dropout_p=embed_dropout_p,
                    gen_aggregation_type=gen_aggregation_type,
                    gen_dropout_p=gen_dropout_p,
                    gen_n_layers=gen_n_layers,
                    gen_attn_hidden=gen_attn_hidden,
                    gen_attn_out=gen_attn_out,
                    gen_d_output=gen_d_output)

        loss = L2Loss()
        output_types = ['prediction']
        super(MATModel, self).__init__(model,
                                       loss=loss,
                                       output_types=output_types,
                                       **kwargs)

    def pad_array(self, array: np.ndarray, shape: Any) -> np.ndarray:
        """
        Pads an array to the desired shape.

        Parameters
        ----------
        array: np.ndarray
        Array to be padded.
        shape: int or Tuple
        Shape the array is padded to.

        Returns
        ----------
        array: np.ndarray
        Array padded to input shape.
        """
        result = np.zeros(shape=shape)
        slices = tuple(slice(s) for s in array.shape)
        result[slices] = array
        return result

    def pad_sequence(self, sequence: np.ndarray) -> np.ndarray:
        """
        Pads a given sequence using the pad_array function.

        Parameters
        ----------
        sequence: np.ndarray
        Arrays in this sequence are padded to the largest shape in the sequence.

        Returns
        ----------
        array: np.ndarray
        Sequence with padded arrays.
        """
        shapes = np.stack([np.array(t.shape) for t in sequence])
        max_shape = tuple(np.max(shapes, axis=0))
        return np.stack([self.pad_array(t, shape=max_shape) for t in sequence])

    def default_generator(self,
                          dataset,
                          epochs=1,
                          mode='fit',
                          deterministic=True,
                          pad_batches=True,
                          **kwargs):
        for epoch in range(epochs):
            for (X_b, y_b, w_b,
                 ids_b) in dataset.iterbatches(batch_size=self.batch_size,
                                               deterministic=deterministic,
                                               pad_batches=pad_batches):

                node_features = self.pad_sequence(
                    [torch.tensor(data.node_features).float() for data in X_b])
                adjacency_matrix = self.pad_sequence([
                    torch.tensor(data.adjacency_matrix).float() for data in X_b
                ])
                distance_matrix = self.pad_sequence([
                    torch.tensor(data.distance_matrix).float() for data in X_b
                ])

                inputs = [node_features, adjacency_matrix, distance_matrix]
                yield (inputs, [y_b], [w_b])



from __future__ import annotations
from typing import Mapping, Tuple, Optional, Union, Iterator, List
import torch
import numpy as np
import warnings
try:
    import pylibxc
except (ImportError, ModuleNotFoundError) as e:
    warnings.warn(
        f"{e}, Failed to import pylibxc. Might not be able to use xc.")

# libxc with derivative

# This is the interface of libxc to pytorch to make the it differentiable
# in pytorch format.
# The torch inputs are flattened and should have been checked to have the
# same length and shape, i.e. (ninps).


class CalcLDALibXCUnpol(torch.autograd.Function):
    """Calculates the energy density or its derivative w.r.t. density for
    unpolarized LDA.

    Local-density approximations (LDA) are a class of approximations to the
    exchange–correlation (XC) energy functional in density functional theory
    (DFT) that depend solely upon the value of the electronic density at each
    point in space (and not, for example, derivatives of the density or the
    Kohn–Sham orbitals).

    The result is a tensor with shape (ninps).

    Examples
    --------
    >>> import torch
    >>> import pylibxc
    >>> libxcfcn = pylibxc.LibXCFunctional("lda_x", "unpolarized")
    >>> rho = torch.tensor([0.1, 0.2, 0.3])
    >>> res = CalcLDALibXCUnpol.apply(rho, 0, libxcfcn)[0]
    >>> print(res)
    tensor([[-0.0343, -0.0864, -0.1483]], dtype=torch.float64)

    """

    @staticmethod
    def forward(ctx, rho: torch.Tensor, deriv: int,  # type: ignore
                libxcfcn: pylibxc.functional.LibXCFunctional) -> \
            Tuple[torch.Tensor, ...]:  # type: ignore
        """Calculates and returns the energy density or its derivative w.r.t.
        density.

        Parameters
        ----------
        rho: torch.Tensor
            Density tensor with shape (ninps)
        deriv: int
            Derivative order. 0 for energy density, 1 for derivative w.r.t.
            density, 2 for second derivative w.r.t. density, etc.
        libxcfcn: pylibxc.functional.LibXCFunctional
            libxc functional to use

        Returns
        -------
        Tuple[torch.Tensor]
            Result is a tensor with shape (ninps)

        """

        inp = {
            "rho": rho.detach().numpy(),
        }
        res = _get_libxc_res(inp, deriv, libxcfcn, family=1, polarized=False)[0]

        ctx.save_for_backward(rho, res)
        ctx.deriv = deriv
        ctx.libxcfcn = libxcfcn
        return (res,)

    @staticmethod
    def backward(
        ctx, *grad_res: torch.Tensor
    ) -> Tuple[Optional[torch.Tensor], ...]:  # type: ignore
        """Calculates the gradient w.r.t. the input rho.

        Parameters
        ----------
        grad_res: torch.Tensor
            Gradient of the result w.r.t. the result itself.

        Returns
        -------
        Tuple[torch.Tensor]
            Gradient w.r.t. the input rho.

        """
        rho, res = ctx.saved_tensors

        dres_drho = CalcLDALibXCUnpol.apply(rho, ctx.deriv + 1, ctx.libxcfcn)[0]
        grad_rho = dres_drho * grad_res[0]
        return (grad_rho, None, None)


class CalcLDALibXCPol(torch.autograd.Function):
    """
    Local-density approximations (LDA) are a class of approximations to the
    exchange–correlation (XC) energy functional in density functional theory
    (DFT) that depend solely upon the value of the electronic density at each
    point in space (and not, for example, derivatives of the density or the
    Kohn–Sham orbitals).

    Examples
    --------
    >>> import torch
    >>> import pylibxc
    >>> libxcfcn = pylibxc.LibXCFunctional("lda_x", "polarized")
    >>> rho_u = torch.tensor([0.1, 0.2, 0.3])
    >>> rho_d = torch.tensor([0.1, 0.2, 0.3])
    >>> res = CalcLDALibXCPol.apply(rho_u, rho_d, 0, libxcfcn)[0]
    >>> print(res)
    tensor([[-0.0864, -0.2177, -0.3738]], dtype=torch.float64)

    """

    @staticmethod
    def forward(
        ctx,
        rho_u: torch.Tensor,
        rho_d: torch.Tensor,
        deriv: int,  # type: ignore
        libxcfcn: pylibxc.functional.LibXCFunctional
    ) -> Tuple[torch.Tensor, ...]:
        """Calculates and returns the energy density or its derivative w.r.t.
        density for polarized LDA.

        Parameters
        ----------
        rho_u: torch.Tensor
            Density tensor for spin-up with shape (ninps)
        rho_d: torch.Tensor
            Density tensor for spin-down with shape (ninps)
        deriv: int
            Derivative order. 0 for energy density, 1 for derivative w.r.t.
            density, 2 for second derivative w.r.t. density, etc.
        libxcfcn: pylibxc.functional.LibXCFunctional
            libxc functional to use

        Returns
        -------
        Tuple[torch.Tensor]
            Result is a tensor with shape (nderiv, ninps) where the first
            dimension indicates the result for derivatives of spin-up and
            spin-down and some of its combination.

        """

        inp = {
            "rho": _pack_input(rho_u, rho_d),
        }
        res = _get_libxc_res(inp, deriv, libxcfcn, family=1, polarized=True)[0]

        ctx.save_for_backward(rho_u, rho_d, res)
        ctx.deriv = deriv
        ctx.libxcfcn = libxcfcn
        return (res,)

    @staticmethod
    def backward(ctx,  # type: ignore
                 *grad_res: torch.Tensor) -> \
            Tuple[Optional[torch.Tensor], ...]:  # type: ignore
        """Calculates the gradient w.r.t. the input rho.

        Parameters
        ----------
        grad_res: torch.Tensor
            Gradient of the result w.r.t. the result itself.

        Returns
        -------
        Tuple[torch.Tensor]
            Gradient w.r.t. the input rho.

        """
        inps = ctx.saved_tensors[:2]
        res = ctx.saved_tensors[2:]  # noqa: F841
        deriv = ctx.deriv
        libxcfcn = ctx.libxcfcn

        derivs = CalcLDALibXCPol.apply(*inps, deriv + 1, libxcfcn)

        # generated by `_generate_spin_list(deriv, ["rho"], [2])`
        if deriv == 0:
            deriv_idxs = [[0], [0]]
            spin_idxs: List[List[Tuple[int, ...]]] = [[(0,)], [(1,)]]
        elif deriv == 1:
            deriv_idxs = [[0], [0]]
            spin_idxs = [[(0, 1)], [(1, 2)]]
        elif deriv == 2:
            deriv_idxs = [[0], [0]]
            spin_idxs = [[(0, 1, 2)], [(1, 2, 3)]]
        elif deriv == 3:
            deriv_idxs = [[0], [0]]
            spin_idxs = [[(0, 1, 2, 3)], [(1, 2, 3, 4)]]
        else:
            raise RuntimeError(
                f"Unimplemented derivative for deriv == {deriv} for polarized LDA"
            )

        grad_inps = _get_grad_inps(grad_res, inps, derivs, ctx.needs_input_grad,
                                   deriv_idxs, spin_idxs)
        return (*grad_inps, None, None)


class CalcGGALibXCUnpol(torch.autograd.Function):
    """Calculates the energy density or its derivative w.r.t. density for
    unpolarized GGA.

    Generalized-gradient approximations (GGA) are a class of approximations to
    the exchange–correlation (XC) energy functional in density functional theory
    (DFT) that depend not only upon the value of the electronic density at each
    point in space, but also upon its gradient.

    Examples
    --------
    >>> import torch
    >>> import pylibxc
    >>> libxcfcn = pylibxc.LibXCFunctional("gga_c_pbe", "unpolarized")
    >>> rho = torch.tensor([0.1, 0.2, 0.3])
    >>> sigma = torch.tensor([0.1, 0.2, 0.3])
    >>> res = CalcGGALibXCUnpol.apply(rho, sigma, 0, libxcfcn)[0]
    >>> print(res)
    tensor([[-0.0016, -0.0070, -0.0137]], dtype=torch.float64)

    """

    @staticmethod
    def forward(ctx, rho: torch.Tensor, sigma: torch.Tensor, deriv: int,  # type: ignore
                libxcfcn: pylibxc.functional.LibXCFunctional) ->\
            Tuple[torch.Tensor, ...]:  # type: ignore
        """Calculates and returns the energy density or its derivative w.r.t.
        density and contracted gradient.

        Every element in the tuple is a tensor with shape (ninps)

        Parameters
        ----------
        rho: torch.Tensor
            Density tensor with shape (ninps)
        sigma: torch.Tensor
            Contracted gradient tensor with shape (ninps)
        deriv: int
            Derivative order. 0 for energy density, 1 for derivative w.r.t.
            density, 2 for second derivative w.r.t. density, etc.
        libxcfcn: pylibxc.functional.LibXCFunctional
            libxc functional to use

        """

        inp = {
            "rho": rho,
            "sigma": sigma,
        }
        # for gga, res is a tuple
        res = _get_libxc_res(inp, deriv, libxcfcn, family=2, polarized=False)

        ctx.save_for_backward(rho, sigma, *res)
        ctx.deriv = deriv
        ctx.libxcfcn = libxcfcn
        return (*res,)

    @staticmethod
    def backward(ctx, *grad_res: torch.Tensor) -> \
            Tuple[Optional[torch.Tensor], ...]:  # type: ignore
        """Calculates the gradient w.r.t. the input rho and sigma.

        Parameters
        ----------
        grad_res : torch.Tensor
            Gradient of the result w.r.t. the result itself.

        Returns
        -------
        Tuple[torch.Tensor]
            Gradient w.r.t. the input rho and sigma.

        """
        inps = ctx.saved_tensors[:2]
        res = ctx.saved_tensors[2:]  # noqa: F841
        deriv = ctx.deriv
        libxcfcn = ctx.libxcfcn

        derivs = CalcGGALibXCUnpol.apply(*inps, deriv + 1, libxcfcn)

        # generated by _generate_pair_deriv_idxs(deriv, ["rho", "sigma"])
        # see _get_grad_inps for explanation about deriv_idxs
        if deriv == 0:
            deriv_idxs = [[0], [1]]
        elif deriv == 1:
            deriv_idxs = [[0, 1], [1, 2]]
        elif deriv == 2:
            deriv_idxs = [[0, 1, 2], [1, 2, 3]]
        elif deriv == 3:
            deriv_idxs = [[0, 1, 2, 3], [1, 2, 3, 4]]
        else:
            raise RuntimeError("Cannot handle GGA deriv %d" % deriv)

        grad_inps = _get_grad_inps(grad_res, inps, derivs, ctx.needs_input_grad,
                                   deriv_idxs)
        return (*grad_inps, None, None)


class CalcGGALibXCPol(torch.autograd.Function):
    """Calculates the energy density or its derivative w.r.t. density for
    polarized GGA.

    Generalized-gradient approximations (GGA) are a class of approximations to
    the exchange–correlation (XC) energy functional in density functional theory
    (DFT) that depend not only upon the value of the electronic density at each
    point in space, but also upon its gradient.

    Examples
    --------
    >>> import torch
    >>> import pylibxc
    >>> libxcfcn = pylibxc.LibXCFunctional("gga_c_pbe", "polarized")
    >>> rho_u = torch.tensor([0.1, 0.2, 0.3])
    >>> rho_d = torch.tensor([0.1, 0.2, 0.3])
    >>> sigma_uu = torch.tensor([0.1, 0.2, 0.3])
    >>> sigma_ud = torch.tensor([0.1, 0.2, 0.3])
    >>> sigma_dd = torch.tensor([0.1, 0.2, 0.3])
    >>> res = CalcGGALibXCPol.apply(rho_u, rho_d, sigma_uu, sigma_ud, sigma_dd, 0, libxcfcn)[0]
    >>> print(res)
    tensor([[-0.0047, -0.0175, -0.0322]], dtype=torch.float64)

    """

    @staticmethod
    def forward(ctx, rho_u: torch.Tensor, rho_d: torch.Tensor,  # type: ignore
                sigma_uu: torch.Tensor, sigma_ud: torch.Tensor, sigma_dd: torch.Tensor,
                deriv: int, libxcfcn: pylibxc.functional.LibXCFunctional) -> \
            Tuple[torch.Tensor, ...]:  # type: ignore
        """
        Calculates and returns the energy density or its derivative w.r.t.
        density and contracted gradient.

        Parameters
        ----------
        rho_u: torch.Tensor
            Density tensor for spin-up with shape (ninps)
        rho_d: torch.Tensor
            Density tensor for spin-down with shape (ninps)
        sigma_uu: torch.Tensor
            Contracted gradient tensor for spin-up with shape (ninps)
        sigma_ud: torch.Tensor
            Contracted gradient tensor for spin-up and spin-down with shape (ninps)
        sigma_dd: torch.Tensor
            Contracted gradient tensor for spin-down with shape (ninps)
        deriv: int
            Derivative order. 0 for energy density, 1 for derivative w.r.t.
            density, 2 for second derivative w.r.t. density, etc.
        libxcfcn: pylibxc.functional.LibXCFunctional
            libxc functional to use

        Returns
        -------
        Tuple[torch.Tensor]
            Result is a tensor with shape (nderiv, ninps) where the first
            dimension indicates the result for derivatives of spin-up and
            spin-down and some of its combination.

        """

        inp = {
            "rho": _pack_input(rho_u, rho_d),
            "sigma": _pack_input(sigma_uu, sigma_ud, sigma_dd),
        }
        res = _get_libxc_res(inp, deriv, libxcfcn, family=2, polarized=True)

        ctx.save_for_backward(rho_u, rho_d, sigma_uu, sigma_ud, sigma_dd, *res)
        ctx.deriv = deriv
        ctx.libxcfcn = libxcfcn
        return (*res,)

    @staticmethod
    def backward(ctx, *grad_res: torch.Tensor) -> \
            Tuple[Optional[torch.Tensor], ...]:  # type: ignore
        """Returns the gradient w.r.t. the input rho and sigma.

        Parameters
        ----------
        grad_res : torch.Tensor
            Gradient of the result w.r.t. the result itself.

        Returns
        -------
        Tuple[torch.Tensor]
            Gradient w.r.t. the input rho and sigma.

        """
        inps = ctx.saved_tensors[:5]
        res = ctx.saved_tensors[5:]  # noqa: F841
        deriv = ctx.deriv
        libxcfcn = ctx.libxcfcn

        derivs = CalcGGALibXCPol.apply(*inps, deriv + 1, libxcfcn)

        # generated by `_generate_spin_list(deriv, ["rho", "sigma"], [2, 3])`
        if deriv == 0:
            deriv_idxs = [[0], [0], [1], [1], [1]]
            spin_idxs: List[List[Tuple[int, ...]]] = [[(0,)], [(1,)], [(0,)],
                                                      [(1,)], [(2,)]]
        elif deriv == 1:
            deriv_idxs = [[0, 1], [0, 1], [1, 2], [1, 2], [1, 2]]
            spin_idxs = [[(0, 1), (0, 1, 2)], [(1, 2), (3, 4, 5)],
                         [(0, 3), (0, 1, 2)], [(1, 4), (1, 3, 4)],
                         [(2, 5), (2, 4, 5)]]
        elif deriv == 2:
            deriv_idxs = [[0, 1, 2], [0, 1, 2], [1, 2, 3], [1, 2, 3], [1, 2, 3]]
            spin_idxs = [[(0, 1, 2), (0, 1, 2, 3, 4, 5), (0, 1, 2, 3, 4, 5)],
                         [(1, 2, 3), (3, 4, 5, 6, 7, 8), (6, 7, 8, 9, 10, 11)],
                         [(0, 3, 6), (0, 1, 2, 6, 7, 8), (0, 1, 2, 3, 4, 5)],
                         [(1, 4, 7), (1, 3, 4, 7, 9, 10), (1, 3, 4, 6, 7, 8)],
                         [(2, 5, 8), (2, 4, 5, 8, 10, 11), (2, 4, 5, 7, 8, 9)]]
        elif deriv == 3:
            deriv_idxs = [[0, 1, 2, 3], [0, 1, 2, 3], [1, 2, 3, 4],
                          [1, 2, 3, 4], [1, 2, 3, 4]]
            spin_idxs = [[(0, 1, 2, 3), (0, 1, 2, 3, 4, 5, 6, 7, 8),
                          (0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11),
                          (0, 1, 2, 3, 4, 5, 6, 7, 8, 9)],
                         [(1, 2, 3, 4), (3, 4, 5, 6, 7, 8, 9, 10, 11),
                          (6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17),
                          (10, 11, 12, 13, 14, 15, 16, 17, 18, 19)],
                         [(0, 3, 6, 9), (0, 1, 2, 6, 7, 8, 12, 13, 14),
                          (0, 1, 2, 3, 4, 5, 10, 11, 12, 13, 14, 15),
                          (0, 1, 2, 3, 4, 5, 6, 7, 8, 9)],
                         [(1, 4, 7, 10), (1, 3, 4, 7, 9, 10, 13, 15, 16),
                          (1, 3, 4, 6, 7, 8, 11, 13, 14, 16, 17, 18),
                          (1, 3, 4, 6, 7, 8, 10, 11, 12, 13)],
                         [(2, 5, 8, 11), (2, 4, 5, 8, 10, 11, 14, 16, 17),
                          (2, 4, 5, 7, 8, 9, 12, 14, 15, 17, 18, 19),
                          (2, 4, 5, 7, 8, 9, 11, 12, 13, 14)]]
        else:
            raise RuntimeError(
                f"Unimplemented derivative for deriv == {deriv} for polarized GGA"
            )

        grad_inps = _get_grad_inps(grad_res, inps, derivs, ctx.needs_input_grad,
                                   deriv_idxs, spin_idxs)
        return (*grad_inps, None, None)


class CalcMGGALibXCUnpol(torch.autograd.Function):
    """Calculates the energy density or its derivative w.r.t. density for
    unpolarized meta-GGA.

    Meta-generalized-gradient approximations (meta-GGA) are a class of approximations
    to the exchange–correlation (XC) energy functional in density functional theory
    (DFT) that depend not only upon the value of the electronic density at each
    point in space and its gradient, but also upon the Laplacian of the density
    and the kinetic energy density.

    Examples
    --------
    >>> import torch
    >>> import pylibxc
    >>> libxcfcn = pylibxc.LibXCFunctional("mgga_c_m06_l", "unpolarized")
    >>> rho = torch.tensor([0.1, 0.2, 0.3])
    >>> sigma = torch.tensor([0.1, 0.2, 0.3])
    >>> lapl = torch.tensor([0.1, 0.2, 0.3])
    >>> kin = torch.tensor([0.1, 0.2, 0.3])
    >>> res = CalcMGGALibXCUnpol.apply(rho, sigma, lapl, kin, 0, libxcfcn)[0]
    >>> print(res)
    tensor([[-0.0032, -0.0066, -0.0087]], dtype=torch.float64)

    """

    @staticmethod
    def forward(ctx, rho: torch.Tensor, sigma: torch.Tensor, lapl: torch.Tensor,  # type: ignore
                kin: torch.Tensor, deriv: int,
                libxcfcn: pylibxc.functional.LibXCFunctional) ->\
            Tuple[torch.Tensor, ...]:  # type: ignore
        """
        Calculates and returns the energy density or its derivative w.r.t.
        density and contracted gradient.

        Parameters
        ----------
        rho: torch.Tensor
            Density tensor with shape (ninps)
        sigma: torch.Tensor
            Contracted gradient tensor with shape (ninps)
        lapl: torch.Tensor
            Laplacian tensor with shape (ninps)
        kin: torch.Tensor
            Kinetic energy density tensor with shape (ninps)
        deriv: int
            Derivative order. 0 for energy density, 1 for derivative w.r.t.
            density, 2 for second derivative w.r.t. density, etc.
        libxcfcn: pylibxc.functional.LibXCFunctional
            libxc functional to use

        Returns
        -------
        Tuple[torch.Tensor]
            The result is a tensor with shape (nderiv, ninps)

        """

        inp = {
            "rho": rho,
            "sigma": sigma,
            "lapl": lapl,
            "tau": kin,
        }
        # res is a tuple
        res = _get_libxc_res(inp, deriv, libxcfcn, family=4, polarized=False)

        ctx.save_for_backward(rho, sigma, lapl, kin, *res)
        ctx.deriv = deriv
        ctx.libxcfcn = libxcfcn
        return (*res,)

    @staticmethod
    def backward(ctx, *grad_res: torch.Tensor) -> \
            Tuple[Optional[torch.Tensor], ...]:  # type: ignore
        """Returns the gradient w.r.t. the inputs

        Parameters
        ----------
        grad_res : torch.Tensor
            The gradient of the result w.r.t. the result itself.

        Returns
        -------
        Tuple[torch.Tensor]
            The gradient w.r.t. the inputs

        """
        inps = ctx.saved_tensors[:4]
        res = ctx.saved_tensors[4:]  # noqa: F841
        deriv = ctx.deriv
        libxcfcn = ctx.libxcfcn

        derivs = CalcMGGALibXCUnpol.apply(*inps, deriv + 1, libxcfcn)

        # generated by _generate_pair_deriv_idxs(deriv, ["rho", "sigma", "lapl", "tau"])
        # see _get_grad_inps for explanation about deriv_idxs
        if deriv == 0:
            deriv_idxs = [[0], [1], [2], [3]]
        elif deriv == 1:
            deriv_idxs = [[0, 1, 2, 3], [1, 4, 5, 6], [2, 5, 7, 8],
                          [3, 6, 8, 9]]
        elif deriv == 2:
            deriv_idxs = [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                          [1, 4, 5, 6, 10, 11, 12, 13, 14, 15],
                          [2, 5, 7, 8, 11, 13, 14, 16, 17, 18],
                          [3, 6, 8, 9, 12, 14, 15, 17, 18, 19]]
        elif deriv == 3:
            deriv_idxs = [[
                0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17,
                18, 19
            ],
                          [
                              1, 4, 5, 6, 10, 11, 12, 13, 14, 15, 20, 21, 22,
                              23, 24, 25, 26, 27, 28, 29
                          ],
                          [
                              2, 5, 7, 8, 11, 13, 14, 16, 17, 18, 21, 23, 24,
                              26, 27, 28, 30, 31, 32, 33
                          ],
                          [
                              3, 6, 8, 9, 12, 14, 15, 17, 18, 19, 22, 24, 25,
                              27, 28, 29, 31, 32, 33, 34
                          ]]
        else:
            raise RuntimeError("Cannot handle MGGA deriv %d" % deriv)

        grad_inps = _get_grad_inps(grad_res, inps, derivs, ctx.needs_input_grad,
                                   deriv_idxs)
        return (*grad_inps, None, None)


class CalcMGGALibXCPol(torch.autograd.Function):
    """Calculates the energy density or its derivative w.r.t. density for
    polarized meta-GGA.

    Meta-generalized-gradient approximations (meta-GGA) are a class of approximations
    to the exchange–correlation (XC) energy functional in density functional theory
    (DFT) that depend not only upon the value of the electronic density at each
    point in space and its gradient, but also upon the Laplacian of the density
    and the kinetic energy density.

    Examples
    --------
    >>> import torch
    >>> import pylibxc
    >>> libxcfcn = pylibxc.LibXCFunctional("mgga_c_m06_l", "polarized")
    >>> rho_u = torch.tensor([0.1, 0.2, 0.3])
    >>> rho_d = torch.tensor([0.1, 0.2, 0.3])
    >>> sigma_uu = torch.tensor([0.1, 0.2, 0.3])
    >>> sigma_ud = torch.tensor([0.1, 0.2, 0.3])
    >>> sigma_dd = torch.tensor([0.1, 0.2, 0.3])
    >>> lapl_u = torch.tensor([0.1, 0.2, 0.3])
    >>> lapl_d = torch.tensor([0.1, 0.2, 0.3])
    >>> kin_u = torch.tensor([0.1, 0.2, 0.3])
    >>> kin_d = torch.tensor([0.1, 0.2, 0.3])
    >>> res = CalcMGGALibXCPol.apply(rho_u, rho_d, sigma_uu, sigma_ud, sigma_dd,
    ...                               lapl_u, lapl_d, kin_u, kin_d, 0, libxcfcn)[0]
    >>> print(res)
    tensor([[-0.0065, -0.0115, -0.0162]], dtype=torch.float64)

    """

    @staticmethod
    def forward(ctx, rho_u: torch.Tensor, rho_d: torch.Tensor,  # type: ignore
                sigma_uu: torch.Tensor, sigma_ud: torch.Tensor, sigma_dd: torch.Tensor,
                lapl_u: torch.Tensor, lapl_d: torch.Tensor,
                kin_u: torch.Tensor, kin_d: torch.Tensor,
                deriv: int, libxcfcn: pylibxc.functional.LibXCFunctional) -> \
            Tuple[torch.Tensor, ...]:  # type: ignore
        """Calculates and returns the energy density or its derivative w.r.t.
        density and contracted gradient and laplacian and kinetic energy density.
        Every element in the tuple is a tensor with shape of (nderiv, ninps)
        where nderiv depends on the number of derivatives for spin-up and
        spin-down combinations, e.g. nderiv == 3 for vsigma (see libxc manual)

        Parameters
        ----------
        rho_u : torch.Tensor
            The density tensor for spin-up with shape (ninps)
        rho_d : torch.Tensor
            The density tensor for spin-down with shape (ninps)
        sigma_uu : torch.Tensor
            The contracted gradient tensor for spin-up with shape (ninps)
        sigma_ud : torch.Tensor
            The contracted gradient tensor for spin-up and spin-down with shape (ninps)
        sigma_dd : torch.Tensor
            The contracted gradient tensor for spin-down with shape (ninps)
        lapl_u : torch.Tensor
            The laplacian tensor for spin-up with shape (ninps)
        lapl_d : torch.Tensor
            The laplacian tensor for spin-down with shape (ninps)
        kin_u : torch.Tensor
            The kinetic energy density tensor for spin-up with shape (ninps)
        kin_d : torch.Tensor
            The kinetic energy density tensor for spin-down with shape (ninps)
        deriv : int
            The derivative order. 0 for energy density, 1 for derivative w.r.t.
            density, 2 for second derivative w.r.t. density, etc.
        libxcfcn : pylibxc.functional.LibXCFunctional
            The libxc functional to use

        Returns
        -------
        Tuple[torch.Tensor]
            The result is a tensor with shape (nderiv, ninps)
            The result is a tensor with shape (nderiv, ninps) where the first
            dimension indicates the result for derivatives of spin-up and
            spin-down and some of its combination.

        """
        inp = {
            "rho": _pack_input(rho_u, rho_d),
            "sigma": _pack_input(sigma_uu, sigma_ud, sigma_dd),
            "lapl": _pack_input(lapl_u, lapl_d),
            "tau": _pack_input(kin_u, kin_d),
        }
        res = _get_libxc_res(inp, deriv, libxcfcn, family=4, polarized=True)

        ctx.save_for_backward(rho_u, rho_d, sigma_uu, sigma_ud, sigma_dd,
                              lapl_u, lapl_d, kin_u, kin_d, *res)
        ctx.deriv = deriv
        ctx.libxcfcn = libxcfcn
        return (*res,)

    @staticmethod
    def backward(ctx, *grad_res: torch.Tensor) -> \
            Tuple[Optional[torch.Tensor], ...]:  # type: ignore
        """Returns the gradient w.r.t. the input rho and sigma.

        Parameters
        ----------
        grad_res : torch.Tensor
            The gradient of the result w.r.t. the result itself.

        Returns
        -------
        Tuple[torch.Tensor]
            The gradient w.r.t. the input rho and sigma.

        """
        inps = ctx.saved_tensors[:9]
        res = ctx.saved_tensors[9:]  # noqa: F841
        deriv = ctx.deriv
        libxcfcn = ctx.libxcfcn

        derivs = CalcMGGALibXCPol.apply(*inps, deriv + 1, libxcfcn)

        # generated by `_generate_spin_list(deriv, ["rho", "sigma"], [2, 3])`
        if deriv == 0:
            deriv_idxs = [[0], [0], [1], [1], [1], [2], [2], [3], [3]]
            spin_idxs: List[List[Tuple[int, ...]]] = [[(0,)], [(1,)], [(0,)],
                                                      [(1,)], [(2,)], [(0,)],
                                                      [(1,)], [(0,)], [(1,)]]
        elif deriv == 1:
            deriv_idxs = [[0, 1, 2, 3], [0, 1, 2, 3], [1, 4, 5,
                                                       6], [1, 4, 5, 6],
                          [1, 4, 5, 6], [2, 5, 7, 8], [2, 5, 7, 8],
                          [3, 6, 8, 9], [3, 6, 8, 9]]
            spin_idxs = [[(0, 1), (0, 1, 2), (0, 1), (0, 1)],
                         [(1, 2), (3, 4, 5), (2, 3), (2, 3)],
                         [(0, 3), (0, 1, 2), (0, 1), (0, 1)],
                         [(1, 4), (1, 3, 4), (2, 3), (2, 3)],
                         [(2, 5), (2, 4, 5), (4, 5), (4, 5)],
                         [(0, 2), (0, 2, 4), (0, 1), (0, 1)],
                         [(1, 3), (1, 3, 5), (1, 2), (2, 3)],
                         [(0, 2), (0, 2, 4), (0, 2), (0, 1)],
                         [(1, 3), (1, 3, 5), (1, 3), (1, 2)]]
        elif deriv == 2:
            deriv_idxs = [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                          [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                          [1, 4, 5, 6, 10, 11, 12, 13, 14, 15],
                          [1, 4, 5, 6, 10, 11, 12, 13, 14, 15],
                          [1, 4, 5, 6, 10, 11, 12, 13, 14, 15],
                          [2, 5, 7, 8, 11, 13, 14, 16, 17, 18],
                          [2, 5, 7, 8, 11, 13, 14, 16, 17, 18],
                          [3, 6, 8, 9, 12, 14, 15, 17, 18, 19],
                          [3, 6, 8, 9, 12, 14, 15, 17, 18, 19]]
            spin_idxs = [[(0, 1, 2), (0, 1, 2, 3, 4, 5), (0, 1, 2, 3),
                          (0, 1, 2, 3), (0, 1, 2, 3, 4, 5), (0, 1, 2, 3, 4, 5),
                          (0, 1, 2, 3, 4, 5), (0, 1, 2), (0, 1, 2, 3),
                          (0, 1, 2)],
                         [(1, 2, 3), (3, 4, 5, 6, 7, 8), (2, 3, 4, 5),
                          (2, 3, 4, 5), (6, 7, 8, 9, 10, 11),
                          (6, 7, 8, 9, 10, 11), (6, 7, 8, 9, 10, 11), (3, 4, 5),
                          (4, 5, 6, 7), (3, 4, 5)],
                         [(0, 3, 6), (0, 1, 2, 6, 7, 8), (0, 1, 6, 7),
                          (0, 1, 6, 7), (0, 1, 2, 3, 4, 5), (0, 1, 2, 3, 4, 5),
                          (0, 1, 2, 3, 4, 5), (0, 1, 2), (0, 1, 2, 3),
                          (0, 1, 2)],
                         [(1, 4, 7), (1, 3, 4, 7, 9, 10), (2, 3, 8, 9),
                          (2, 3, 8, 9), (1, 3, 4, 6, 7, 8), (2, 3, 6, 7, 8, 9),
                          (2, 3, 6, 7, 8, 9), (3, 4, 5), (4, 5, 6, 7),
                          (3, 4, 5)],
                         [(2, 5, 8), (2, 4, 5, 8, 10, 11), (4, 5, 10, 11),
                          (4, 5, 10, 11), (2, 4, 5, 7, 8, 9),
                          (4, 5, 8, 9, 10, 11), (4, 5, 8, 9, 10, 11), (6, 7, 8),
                          (8, 9, 10, 11), (6, 7, 8)],
                         [(0, 2, 4), (0, 2, 4, 6, 8, 10), (0, 1, 3, 4),
                          (0, 1, 4, 5), (0, 2, 4, 6, 8, 10), (0, 1, 3, 4, 6, 7),
                          (0, 1, 4, 5, 8, 9), (0, 1, 2), (0, 1, 2, 3),
                          (0, 1, 2)],
                         [(1, 3, 5), (1, 3, 5, 7, 9, 11), (1, 2, 4, 5),
                          (2, 3, 6, 7), (1, 3, 5, 7, 9, 11), (1, 2, 4, 5, 7, 8),
                          (2, 3, 6, 7, 10, 11), (1, 2, 3), (2, 3, 4, 5),
                          (3, 4, 5)],
                         [(0, 2, 4), (0, 2, 4, 6, 8, 10), (0, 2, 4, 6),
                          (0, 1, 3, 4), (0, 2, 4, 6, 8, 10),
                          (0, 2, 4, 6, 8, 10), (0, 1, 3, 4, 6, 7), (0, 2, 4),
                          (0, 1, 3, 4), (0, 1, 2)],
                         [(1, 3, 5), (1, 3, 5, 7, 9, 11), (1, 3, 5, 7),
                          (1, 2, 4, 5), (1, 3, 5, 7, 9, 11),
                          (1, 3, 5, 7, 9, 11), (1, 2, 4, 5, 7, 8), (1, 3, 5),
                          (1, 2, 4, 5), (1, 2, 3)]]
        else:
            raise RuntimeError(
                f"Unimplemented derivative for deriv == {deriv} for polarized MGGA"
            )

        grad_inps = _get_grad_inps(grad_res, inps, derivs, ctx.needs_input_grad,
                                   deriv_idxs, spin_idxs)
        return (*grad_inps, None, None)


def _get_libxc_res(inp: Mapping[str, Union[np.ndarray, Tuple[np.ndarray, ...],
                                           torch.Tensor, Tuple[torch.Tensor,
                                                               ...]]],
                   deriv: int, libxcfcn: pylibxc.functional.LibXCFunctional,
                   family: int, polarized: bool) -> Tuple[torch.Tensor, ...]:
    """
    Calls the libxc functional to calculate the energy density or its derivative
    w.r.t. density and contracted gradient.

    Examples
    --------
    >>> import torch
    >>> import pylibxc
    >>> libxcfcn = pylibxc.LibXCFunctional("lda_x", "unpolarized")
    >>> rho = torch.tensor([0.1, 0.2, 0.3])
    >>> res = _get_libxc_res({"rho": rho}, 0, libxcfcn, 2, False)
    >>> print(res)
    (tensor([[-0.0343, -0.0864, -0.1483]], dtype=torch.float64),)

    Parameters
    ----------
    inp: Mapping[str, Union[np.ndarray, Tuple[np.ndarray, ...], torch.Tensor, Tuple[torch.Tensor, ...]]]
        Input data for the libxc functional
    deriv: int
        Derivative order. 0 for energy density, 1 for derivative w.r.t.
        density, 2 for second derivative w.r.t. density, etc.
    libxcfcn: pylibxc.functional.LibXCFunctional
        libxc functional to use
    family: int
        Family of the functional. 2 for GGA and 4 for MGGA
    polarized: bool
        Whether the calculation is for polarized or unpolarized data

    Returns
    -------
    Tuple[torch.Tensor]
        Result is a tuple of tensors with shape (ninps)
        where the first element is the result for the energy density or its
        derivative w.r.t. density and the rest are the contracted gradient.

    """
    do_exc, do_vxc, do_fxc, do_kxc, do_lxc = _get_dos(deriv)

    res = libxcfcn.compute(inp,
                           do_exc=do_exc,
                           do_vxc=do_vxc,
                           do_fxc=do_fxc,
                           do_kxc=do_kxc,
                           do_lxc=do_lxc)

    # compile the results in a tuple with order given in the *_KEYS (e.g. LDA_KEYS)
    res = _extract_returns(res, deriv, family)

    # In libxc, "zk" is the only one returning the energy density
    # per unit volume PER UNIT PARTICLE.
    # everything else is represented by the energy density per unit volume
    # only.
    if deriv == 0:
        rho = inp["rho"]
        if polarized:
            assert isinstance(rho, np.ndarray)
            start = np.zeros(1, dtype=rho.dtype)
            rho = sum(_unpack_input(rho), start)  # rho[:, 0] + rho[:, 1]
        res0 = res[0] * rho
        res = (res0, *res[1:])

    return res


def _pack_input(*vals: torch.Tensor) -> np.ndarray:
    """Arrange the values in a numpy array with fortran memory order

    Examples
    --------
    >>> rho = torch.tensor([[1, 2], [3, 4]])
    >>> sigma = torch.tensor([[1, 2], [3, 4]])
    >>> _pack_input(rho, sigma)
    array([[[1, 1],
            [3, 3]],
    <BLANKLINE>
           [[2, 2],
            [4, 4]]])

    Parameters
    ----------
    vals: torch.Tensor
        Input values

    Returns
    -------
    np.ndarray
        Input values in a numpy array with fortran memory order

    """
    vals_np = np.asarray([val.detach().numpy() for val in vals])
    return np.ascontiguousarray(vals_np.T)


def _unpack_input(inp: np.ndarray) -> Iterator[np.ndarray]:
    """unpack from libxc input format into tuple of inputs

    Examples
    --------
    >>> inp = np.array([[1, 3], [2, 4]])
    >>> tuple(_unpack_input(inp))
    (array([1, 2]), array([3, 4]))

    Parameters
    ----------
    inp: np.ndarray
        Input values in a numpy array with fortran memory order

    Returns
    -------
    Iterator[np.ndarray]
        Unpacked input values

    """
    return (a for a in inp.T)


def _get_dos(deriv: int) -> Tuple[bool, ...]:
    """get the boolean flags for the derivatives

    Examples
    --------
    >>> _get_dos(0)
    (True, False, False, False, False)
    >>> _get_dos(1)
    (False, True, False, False, False)

    Parameters
    ----------
    deriv: int
        Derivative order. 0 for energy density, 1 for derivative w.r.t.
        density, 2 for second derivative w.r.t. density, etc.

    Returns
    -------
    Tuple[bool]
        Boolean flags for the derivatives

    """
    do_exc = deriv == 0
    do_vxc = deriv == 1
    do_fxc = deriv == 2
    do_kxc = deriv == 3
    do_lxc = deriv == 4
    return do_exc, do_vxc, do_fxc, do_kxc, do_lxc


# generated by [_generate_keys(i, ["rho", "sigma"]) for i in range(5)]
# _generate_keys function is below
LDA_KEYS = [["zk"], ["vrho"], ["v2rho2"], ["v3rho3"], ["v4rho4"]]
GGA_KEYS = [["zk"], ["vrho", "vsigma"], ["v2rho2", "v2rhosigma", "v2sigma2"],
            ["v3rho3", "v3rho2sigma", "v3rhosigma2", "v3sigma3"],
            [
                "v4rho4", "v4rho3sigma", "v4rho2sigma2", "v4rhosigma3",
                "v4sigma4"
            ]]
MGGA_KEYS = [['zk'], ['vrho', 'vsigma', 'vlapl', 'vtau'],
             [
                 'v2rho2', 'v2rhosigma', 'v2rholapl', 'v2rhotau', 'v2sigma2',
                 'v2sigmalapl', 'v2sigmatau', 'v2lapl2', 'v2lapltau', 'v2tau2'
             ],
             [
                 'v3rho3', 'v3rho2sigma', 'v3rho2lapl', 'v3rho2tau',
                 'v3rhosigma2', 'v3rhosigmalapl', 'v3rhosigmatau', 'v3rholapl2',
                 'v3rholapltau', 'v3rhotau2', 'v3sigma3', 'v3sigma2lapl',
                 'v3sigma2tau', 'v3sigmalapl2', 'v3sigmalapltau', 'v3sigmatau2',
                 'v3lapl3', 'v3lapl2tau', 'v3lapltau2', 'v3tau3'
             ],
             [
                 'v4rho4', 'v4rho3sigma', 'v4rho3lapl', 'v4rho3tau',
                 'v4rho2sigma2', 'v4rho2sigmalapl', 'v4rho2sigmatau',
                 'v4rho2lapl2', 'v4rho2lapltau', 'v4rho2tau2', 'v4rhosigma3',
                 'v4rhosigma2lapl', 'v4rhosigma2tau', 'v4rhosigmalapl2',
                 'v4rhosigmalapltau', 'v4rhosigmatau2', 'v4rholapl3',
                 'v4rholapl2tau', 'v4rholapltau2', 'v4rhotau3', 'v4sigma4',
                 'v4sigma3lapl', 'v4sigma3tau', 'v4sigma2lapl2',
                 'v4sigma2lapltau', 'v4sigma2tau2', 'v4sigmalapl3',
                 'v4sigmalapl2tau', 'v4sigmalapltau2', 'v4sigmatau3', 'v4lapl4',
                 'v4lapl3tau', 'v4lapl2tau2', 'v4lapltau3', 'v4tau4'
             ]]


def _extract_returns(ret: Mapping[str, np.ndarray], deriv: int, family: int) -> \
        Tuple[torch.Tensor, ...]:
    """Compile the returns from pylibxc into a tuple of tensors with order given
    by the keys

    Examples
    --------
    >>> import numpy as np
    >>> ret = {"zk": np.array([1, 2, 3])}
    >>> _extract_returns(ret, 0, 1)
    (tensor([1, 2, 3]),)

    Parameters
    ----------
    ret: Mapping[str, np.ndarray]
        Return values from pylibxc
    deriv: int
        Derivative order. 0 for energy density, 1 for derivative w.r.t.
        density, 2 for second derivative w.r.t. density, etc.
    family: int
        Family of the functional. 1 for LDA, 2 for GGA, 4 for MGGA

    Returns
    -------
    Tuple[torch.Tensor]
        Result is a tuple of tensors with shape (ninps)
        where the first element is the result for the energy density or its
        derivative w.r.t. density and the rest are the contracted gradient.

    """

    def a(v):
        return torch.as_tensor(v.T)

    if family == 1:
        keys = LDA_KEYS
    elif family == 2:
        keys = GGA_KEYS
    elif family == 4:
        keys = MGGA_KEYS
    else:
        raise RuntimeError("Unknown libxc family %d" % family)
    return tuple(a(ret[key]) for key in keys[deriv])


def _get_grad_inps(
    grad_res: Tuple[torch.Tensor, ...],
    inps: Tuple[torch.Tensor, ...],
    derivs: Tuple[torch.Tensor, ...],
    needs_input_grad: List[bool],
    deriv_idxs: List[List[int]],
    spin_idxs: Optional[List[List[Tuple[int, ...]]]] = None
) -> Tuple[Optional[torch.Tensor], ...]:
    """Calculate the grad_inp from grad_res and given deriv_idxs
    each row indicates the input, while the column indicates the index in out
    deriv_idxs[i][j] means that grad_inp[i] += grad_res[j] * derivs[deriv_idxs[i][j]]

    Examples
    --------
    >>> grad_res = [torch.tensor([1, 2, 3]), torch.tensor([4, 5, 6])]
    >>> inps = [torch.tensor([1, 2, 3]), torch.tensor([4, 5, 6])]
    >>> derivs = [torch.tensor([1, 2, 3]), torch.tensor([4, 5, 6])]
    >>> needs_input_grad = [True, True]
    >>> deriv_idxs = [[0], [1]]
    >>> _get_grad_inps(grad_res, inps, derivs, needs_input_grad, deriv_idxs)
    (tensor([1, 4, 9]), tensor([ 4, 10, 18]))

    Parameters
    ----------
    grad_res: Tuple[torch.Tensor]
        Gradient of the result w.r.t. the result itself.
    inps: Tuple[torch.Tensor]
        Input tensors
    derivs: Tuple[torch.Tensor]
        Derivative tensors
    needs_input_grad: List[bool]
        Boolean list indicating whether the input requires grad
    deriv_idxs: List[List[int]]
        List of indices for the derivatives
    spin_idxs: Optional[List[List[Tuple[int, ...]]], optional
        List of spin indices, by default None

    Returns
    -------
    Tuple[Optional[torch.Tensor]]
        Gradient w.r.t. the inputs

    """
    grad_inps: List[Optional[torch.Tensor]] = []
    for i in range(len(deriv_idxs)):
        # if the input does not requires grad, then don't compute
        if not needs_input_grad[i]:
            grad_inps.append(None)
            continue

        grad_inp = torch.zeros_like(inps[i])
        didxs = deriv_idxs[i]
        if spin_idxs is not None:
            sidxs = spin_idxs[i]
        for j in range(len(didxs)):
            if spin_idxs is None:
                grad_inp = grad_inp + grad_res[j] * derivs[didxs[j]]
            else:
                grad_inp = grad_inp + torch.sum(
                    grad_res[j] * derivs[didxs[j]][sidxs[j], :], dim=0)
        grad_inps.append(grad_inp)
    return tuple(grad_inps)



"""Interface for reinforcement learning."""
try:
    from deepchem.rl.a2c import A2C  # noqa: F401
    from deepchem.rl.ppo import PPO  # noqa: F401
except ModuleNotFoundError:
    pass


class Environment(object):
    """An environment in which an actor performs actions to accomplish a task.

    An environment has a current state, which is represented as either a single NumPy
    array, or optionally a list of NumPy arrays.  When an action is taken, that causes
    the state to be updated.  The environment also computes a reward for each action,
    and reports when the task has been terminated (meaning that no more actions may
    be taken).

    Two types of actions are supported.  For environments with discrete action spaces,
    the action is an integer specifying the index of the action to perform (out of a
    fixed list of possible actions).  For environments with continuous action spaces,
    the action is a NumPy array.

    Environment objects should be written to support pickle and deepcopy operations.
    Many algorithms involve creating multiple copies of the Environment, possibly
    running in different processes or even on different computers.
    """

    def __init__(self,
                 state_shape,
                 n_actions=None,
                 state_dtype=None,
                 action_shape=None):
        """Subclasses should call the superclass constructor in addition to doing their own initialization.

        A value should be provided for either n_actions (for discrete action spaces)
        or action_shape (for continuous action spaces), but not both.

        Parameters
        ----------
        state_shape: tuple or list of tuples
            the shape(s) of the array(s) making up the state
        n_actions: int
            the number of discrete actions that can be performed.  If the action space
            is continuous, this should be None.
        state_dtype: dtype or list of dtypes
            the type(s) of the array(s) making up the state.  If this is None, all
            arrays are assumed to be float32.
        action_shape: tuple
            the shape of the array describing an action.  If the action space
            is discrete, this should be none.
        """
        self._state_shape = state_shape
        self._n_actions = n_actions
        self._action_shape = action_shape
        self._state = None
        self._terminated = None
        if state_dtype is None:
            # Assume all arrays are float32.
            import numpy
            try:
                from collections.abc import Sequence as SequenceCollection
            except:
                from collections import Sequence as SequenceCollection
            if isinstance(state_shape[0], SequenceCollection):
                self._state_dtype = [numpy.float32] * len(state_shape)
            else:
                self._state_dtype = numpy.float32
        else:
            self._state_dtype = state_dtype

    @property
    def state(self):
        """The current state of the environment, represented as either a NumPy array or list of arrays.

        If reset() has not yet been called at least once, this is undefined.
        """
        return self._state

    @property
    def terminated(self):
        """Whether the task has reached its end.

        If reset() has not yet been called at least once, this is undefined.
        """
        return self._terminated

    @property
    def state_shape(self):
        """The shape of the arrays that describe a state.

        If the state is a single array, this returns a tuple giving the shape of that array.
        If the state is a list of arrays, this returns a list of tuples where each tuple is
        the shape of one array.
        """
        return self._state_shape

    @property
    def state_dtype(self):
        """The dtypes of the arrays that describe a state.

        If the state is a single array, this returns the dtype of that array.  If the state
        is a list of arrays, this returns a list containing the dtypes of the arrays.
        """
        return self._state_dtype

    @property
    def n_actions(self):
        """The number of possible actions that can be performed in this Environment.

        If the environment uses a continuous action space, this returns None.
        """
        return self._n_actions

    @property
    def action_shape(self):
        """The expected shape of NumPy arrays representing actions.

        If the environment uses a discrete action space, this returns None.
        """
        return self._action_shape

    def reset(self):
        """Initialize the environment in preparation for doing calculations with it.

        This must be called before calling step() or querying the state.  You can call it
        again later to reset the environment back to its original state.
        """
        raise NotImplementedError("Subclasses must implement this")

    def step(self, action):
        """Take a time step by performing an action.

        This causes the "state" and "terminated" properties to be updated.

        Parameters
        ----------
        action: object
            an object describing the action to take

        Returns
        -------
        the reward earned by taking the action, represented as a floating point number
        (higher values are better)
        """
        raise NotImplementedError("Subclasses must implement this")


class GymEnvironment(Environment):
    """This is a convenience class for working with environments from OpenAI Gym."""

    def __init__(self, name):
        """Create an Environment wrapping the OpenAI Gym environment with a specified name."""
        import gym
        self.env = gym.make(name, render_mode='human')
        self.name = name
        space = self.env.action_space
        if 'n' in dir(space):
            super(GymEnvironment,
                  self).__init__(self.env.observation_space.shape, space.n)
        else:
            super(GymEnvironment,
                  self).__init__(self.env.observation_space.shape,
                                 action_shape=space.shape)

    def reset(self):
        self._state = self.env.reset()[0]
        self._terminated = False

    def step(self, action):
        self._state, reward, self._terminated, bool_val, info = self.env.step(
            action)
        return reward

    def __deepcopy__(self, memo):
        return GymEnvironment(self.name)


class Policy(object):
    """A policy for taking actions within an environment.

    A policy is defined by a tf.keras.Model that takes the current state as input
    and performs the necessary calculations.  There are many algorithms for
    reinforcement learning, and they differ in what values they require a policy to
    compute.  That makes it impossible to define a single interface allowing any
    policy to be optimized with any algorithm.  Instead, this interface just tries
    to be as flexible and generic as possible.  Each algorithm must document what
    values it expects the model to output.

    Special handling is needed for models that include recurrent layers.  In that
    case, the model has its own internal state which the learning algorithm must
    be able to specify and query.  To support this, the Policy must do three things:

    1. The Model must take additional inputs that specify the initial states of
        all its recurrent layers.  These will be appended to the list of arrays
        specifying the environment state.

    2. The Model must also return the final states of all its recurrent layers as
        outputs.

    3. The constructor argument rnn_initial_states must be specified to define
        the states to use for the Model's recurrent layers at the start of a new
        rollout.

    Policy objects should be written to support pickling.  Many algorithms involve
    creating multiple copies of the Policy, possibly running in different processes
    or even on different computers.
    """

    def __init__(self, output_names, rnn_initial_states=[]):
        """Subclasses should call the superclass constructor in addition to doing
        their own initialization.

        Parameters
        ----------
        output_names: list of strings
            the names of the Model's outputs, in order.  It is up to each reinforcement
            learning algorithm to document what outputs it expects policies to compute.
            Outputs that return the final states of recurrent layers should have the
            name 'rnn_state'.
        rnn_initial_states: list of NumPy arrays
            the initial states of the Model's recurrent layers at the start of a new
            rollout
        """
        self.output_names = output_names
        self.rnn_initial_states = rnn_initial_states

    def create_model(self, **kwargs):
        """Construct and return a tf.keras.Model that computes the policy.

        The inputs to the model consist of the arrays representing the current state
        of the environment, followed by the initial states for all recurrent layers.
        Depending on the algorithm being used, other inputs might get passed as
        well.  It is up to each algorithm to document that.
        """
        raise NotImplementedError("Subclasses must implement this")



"""
Gradient Boosting Decision Tree wrapper interface
"""

import os
import logging
import tempfile
import warnings
from typing import Callable, Optional, Union, List, Any

import numpy as np
from sklearn.base import BaseEstimator
from sklearn.model_selection import train_test_split

from deepchem.data import Dataset
from deepchem.models.sklearn_models import SklearnModel

logger = logging.getLogger(__name__)


class GBDTModel(SklearnModel):
    """Wrapper class that wraps GBDT models as DeepChem models.

    This class supports LightGBM/XGBoost models.
    """

    def __init__(self,
                 model: BaseEstimator,
                 model_dir: Optional[str] = None,
                 early_stopping_rounds: int = 50,
                 eval_metric: Optional[Union[str, Callable]] = None,
                 **kwargs):
        """
        Parameters
        ----------
        model: BaseEstimator
            The model instance of scikit-learn wrapper LightGBM/XGBoost models.
        model_dir: str, optional (default None)
            Path to directory where model will be stored.
        early_stopping_rounds: int, optional (default 50)
            Activates early stopping. Validation metric needs to improve at least once
            in every early_stopping_rounds round(s) to continue training.
        eval_metric: Union[str, Callable]
            If string, it should be a built-in evaluation metric to use.
            If callable, it should be a custom evaluation metric, see official note for more details.
        """

        try:
            import xgboost
            import lightgbm
        except ModuleNotFoundError:
            raise ImportError(
                'This function requires XGBoost and LightGBM modules '
                'to be installed.')

        if model_dir is not None:
            if not os.path.exists(model_dir):
                os.makedirs(model_dir)
        else:
            model_dir = tempfile.mkdtemp()
        self.model_dir = model_dir
        self.model = model
        self.model_class = model.__class__
        self.early_stopping_rounds = early_stopping_rounds
        self.model_type = self._check_model_type()
        self.callbacks: List[Union[Any, Any]]

        if self.early_stopping_rounds <= 0:
            raise ValueError("Early Stopping Rounds cannot be less than 1.")

        if self.model.__class__.__name__.startswith('XGB'):
            self.callbacks = [
                xgboost.callback.EarlyStopping(
                    rounds=self.early_stopping_rounds)
            ]
        elif self.model.__class__.__name__.startswith('LGBM'):
            self.callbacks = [
                lightgbm.early_stopping(
                    stopping_rounds=self.early_stopping_rounds),
            ]

        if eval_metric is None:
            if self.model_type == "classification":
                self.eval_metric: Optional[Union[str, Callable]] = "auc"
            elif self.model_type == "regression":
                self.eval_metric = "mae"
            else:
                self.eval_metric = eval_metric
        else:
            self.eval_metric = eval_metric

    def _check_model_type(self) -> str:
        class_name = self.model.__class__.__name__
        if class_name.endswith("Classifier"):
            return "classification"
        elif class_name.endswith("Regressor"):
            return "regression"
        elif class_name == "NoneType":
            return "none"
        else:
            raise ValueError(
                "{} is not a supported model instance.".format(class_name))

    def fit(self, dataset: Dataset):
        """Fits GDBT model with all data.

        First, this function splits all data into train and valid data (8:2),
        and finds the best n_estimators. And then, we retrain all data using
        best n_estimators * 1.25.

        Parameters
        ----------
        dataset: Dataset
            The `Dataset` to train this model on.
        """
        X = dataset.X
        y = np.squeeze(dataset.y)

        # GDBT doesn't support multi-output(task)
        if len(y.shape) != 1:
            raise ValueError("GDBT model doesn't support multi-output(task)")

        seed = self.model.random_state
        stratify = None
        if self.model_type == "classification":
            stratify = y

        # Find optimal n_estimators based on original learning_rate and early_stopping_rounds
        X_train, X_test, y_train, y_test = train_test_split(X,
                                                            y,
                                                            test_size=0.2,
                                                            random_state=seed,
                                                            stratify=stratify)
        self.model.fit(
            X_train,
            y_train,
            callbacks=self.callbacks,
            eval_metric=self.eval_metric,
            eval_set=[(X_test, y_test)],
        )

        # retrain model to whole data using best n_estimators * 1.25
        if self.model.__class__.__name__.startswith('XGB'):
            estimated_best_round = np.round(
                (self.model.best_iteration + 1) * 1.25)
        else:
            estimated_best_round = np.round(self.model.best_iteration_ * 1.25)

        self.model.n_estimators = np.int64(estimated_best_round)
        self.model.fit(X, y, eval_metric=self.eval_metric)

    def fit_with_eval(self, train_dataset: Dataset, valid_dataset: Dataset):
        """Fits GDBT model with valid data.

        Parameters
        ----------
        train_dataset: Dataset
            The `Dataset` to train this model on.
        valid_dataset: Dataset
            The `Dataset` to validate this model on.
        """
        X_train, X_valid = train_dataset.X, valid_dataset.X
        y_train, y_valid = np.squeeze(train_dataset.y), np.squeeze(
            valid_dataset.y)

        # GDBT doesn't support multi-output(task)
        if len(y_train.shape) != 1 or len(y_valid.shape) != 1:
            raise ValueError("GDBT model doesn't support multi-output(task)")

        self.model.fit(
            X_train,
            y_train,
            callbacks=self.callbacks,
            eval_metric=self.eval_metric,
            eval_set=[(X_valid, y_valid)],
        )


#########################################
# Deprecation warnings for XGBoostModel
#########################################


class XGBoostModel(GBDTModel):

    def __init__(self, *args, **kwargs):
        warnings.warn(
            "XGBoostModel is deprecated and has been renamed to GBDTModel.",
            FutureWarning,
        )
        super(XGBoostModel, self).__init__(*args, **kwargs)



import scipy.stats as st
import scipy
import unittest


class TestVoxelUtils(unittest.TestCase):

    def test_gibrat(self):
        """
        Test fix of function name 'gilbrat' to 'gibrat' of scipy.stats
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.gilbrat.html
        """
        assert isinstance(st.gibrat, scipy.stats._continuous_distns.gibrat_gen)



""" This module contains the implementation of the various flow layers and models"""
import numpy as np
import torch
import torch.nn as nn
from typing import Optional, Tuple, Union


class Flow(nn.Module):
    """
    Generic class for flow functions

    Flows [flow1]_ should satisfy several conditions in order to be practical. They should:

    - be invertible; for sampling we need g while for computing likelihood we need `f` ,
    - be sufficiently expressive to model the distribution of interest,
    - be computationally efficient, both in terms of computing `f` and `g` (depending on the application) but also in terms of the calculation of the determinant of the Jacobian.

    Flow layers are generally used as a part of a Normalizing Flow model,
    which is a generative model that learns a target distribution by transforming a
    simple base distribution through a series of invertible transformations.
    The target distribution is then defined as the composition of the base distribution
    and the flow transformations.

    References
    ----------
    .. [flow1] Kobyzev, I., Prince, S. J., & Brubaker, M. A. (2020).
        Normalizing flows: An introduction and review of current methods.
        IEEE transactions on pattern analysis and machine intelligence, 43(11), 3964-3979.
    """

    def __init__(self):
        """Initializes the flow function
        """
        super().__init__()

    def forward(self, z: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Forward pass of the flow

        Parameters
        ----------
        z: torch.Tensor
            Input tensor

        Returns
        -------
        z_: torch.Tensor
            Transformed tensor
        log_det: torch.Tensor
            Logarithm of the determinant of the Jacobian of the transformation
        """
        raise NotImplementedError("Forward pass has not been implemented.")

    def inverse(self, z: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Inverse pass of the flow

        Parameters
        ----------
        z: torch.Tensor
            Input tensor

        Returns
        -------
        z_: torch.Tensor
            Transformed tensor
        log_det: torch.Tensor
            Logarithm of the determinant of the Jacobian of the transformation
        """
        raise NotImplementedError("This flow has no algebraic inverse.")


class Affine(Flow):
    """Class which performs the Affine transformation.

    This transformation is based on the affinity of the base distribution with
    the target distribution. A geometric transformation is applied where
    the parameters performs changes on the scale and shift of a function
    (inputs).

    Normalizing Flow transformations must be bijective in order to compute
    the logarithm of jacobian's determinant. For this reason, transformations
    must perform a forward and inverse pass.

    Example
    --------
    >>> import deepchem as dc
    >>> from deepchem.models.torch_models.layers import Affine
    >>> import torch
    >>> from torch.distributions import MultivariateNormal
    >>> # initialize the transformation layer's parameters
    >>> dim = 2
    >>> samples = 96
    >>> transforms = Affine(dim)
    >>> # forward pass based on a given distribution
    >>> distribution = MultivariateNormal(torch.zeros(dim), torch.eye(dim))
    >>> input = distribution.sample(torch.Size((samples, dim)))
    >>> len(transforms.forward(input))
    2
    >>> # inverse pass based on a distribution
    >>> len(transforms.inverse(input))
    2

    """

    def __init__(self, dim: int) -> None:
        """Create a Affine transform layer.

        Parameters
        ----------
        dim: int
            Value of the Nth dimension of the dataset.

        """

        super().__init__()
        self.dim = dim
        self.scale = nn.Parameter(torch.zeros(self.dim))
        self.shift = nn.Parameter(torch.zeros(self.dim))
        self.batch_dims = torch.nonzero(torch.tensor(self.scale.shape) == 1,
                                        as_tuple=False)[:, 0].tolist()

    def forward(self, x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """Performs a transformation between two different distributions. This
        particular transformation represents the following function:

        .. math:: y = x * exp(a) + b

        where a is scale parameter and b performs a shift.
        This class also returns the logarithm of the jacobians determinant
        which is useful when invert a transformation and compute the
        probability of the transformation.

        Parameters
        ----------
        x : torch.Tensor
            Tensor sample with the initial distribution data which will pass into
            the normalizing flow algorithm.

        Returns
        -------
        y : torch.Tensor
            Transformed tensor according to Affine layer with the shape of 'x'.
        log_det_jacobian : torch.Tensor
            Tensor which represents the info about the deviation of the initial
            and target distribution.

        """

        y = torch.exp(self.scale) * x + self.shift
        det_jacobian = torch.exp(self.scale.sum())
        log_det_jacobian = torch.ones(y.shape[0]) * torch.log(det_jacobian)

        return y, log_det_jacobian

    def inverse(self, y: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """Performs a transformation between two different distributions.
        This transformation represents the bacward pass of the function
        mention before. Its mathematical representation is x = (y - b) / exp(a)
        , where "a" is scale parameter and "b" performs a shift. This class
        also returns the logarithm of the jacobians determinant which is
        useful when invert a transformation and compute the probability of
        the transformation.

        Parameters
        ----------
        y : torch.Tensor
            Tensor sample with transformed distribution data which will be used in
            the normalizing algorithm inverse pass.

        Returns
        -------
        x : torch.Tensor
            Transformed tensor according to Affine layer with the shape of 'y'.
        inverse_log_det_jacobian : torch.Tensor
            Tensor which represents the information of the deviation of the initial
            and target distribution.

        """

        x = (y - self.shift) / torch.exp(self.scale)
        det_jacobian = 1 / torch.exp(self.scale.sum())
        inverse_log_det_jacobian = torch.ones(
            x.shape[0]) * torch.log(det_jacobian)

        return x, inverse_log_det_jacobian


class MaskedAffineFlow(Flow):
    """
    This class implements the Masked Affine Flow layer

    The Masked Affine Flow [maskedaffine1]_ layer is a type of normalizing flow layer which
    is used to learn a target distribution. The layer is based on the
    affine flow layer, but with a mask applied to the input data. The mask
    is a tensor of the same size as the input data, filled with 0s and 1s.
    The mask is used to determine which features are transformed by the
    affine flow layer. The affine flow layer is defined as follows:

    Masked affine flow
    .. math:: f(z) = b * z + (1 - b) * (z * e^{s(b * z)} + t)

    Example
    -------
    >>> import torch
    >>> import torch.nn as nn
    >>> import torch.nn.functional as F
    >>> from deepchem.models.torch_models.flows import MaskedAffineFlow
    >>> from torch.distributions import MultivariateNormal

    >>> dim = 2
    >>> samples = 96
    >>> data = MultivariateNormal(torch.zeros(dim), torch.eye(dim))
    >>> tensor = data.sample(torch.Size((samples, dim)))

    >>> layers = 4
    >>> hidden_size = 16
    >>> masks = F.one_hot(torch.tensor([i % 2 for i in range(layers)])).float()

    >>> s_func = nn.Sequential(
    ...     nn.Linear(in_features=dim, out_features=hidden_size), nn.LeakyReLU(),
    ...     nn.Linear(in_features=hidden_size, out_features=hidden_size),
    ...     nn.LeakyReLU(), nn.Linear(in_features=hidden_size, out_features=dim))

    >>> t_func = nn.Sequential(
    ...     nn.Linear(in_features=dim, out_features=hidden_size), nn.LeakyReLU(),
    ...     nn.Linear(in_features=hidden_size, out_features=hidden_size),
    ...     nn.LeakyReLU(), nn.Linear(in_features=hidden_size, out_features=dim))

    >>> layers = nn.ModuleList(
    ...     [MaskedAffineFlow(mask, s_func, t_func) for mask in masks])

    >>> for layer in layers:
    ...   _, inverse_log_det_jacobian = layer.inverse(tensor)
    ...   inverse_log_det_jacobian = inverse_log_det_jacobian.detach().numpy()
    >>> len(inverse_log_det_jacobian)
    96

    References
    ----------
    .. [maskedaffine1] Dinh, L., Sohl-Dickstein, J., & Bengio, S. (2016).
        Density estimation using real nvp. arXiv preprint arXiv:1605.08803.
    """

    def __init__(
        self,
        b: torch.Tensor,
        t: Optional[Union[torch.nn.ModuleList, torch.nn.Sequential]] = None,
        s: Optional[Union[torch.nn.ModuleList, torch.nn.Sequential]] = None
    ) -> None:
        """
        Initializes the Masked Affine Flow layer

        Parameters
        ----------
        b: torch.Tensor
            mask for features, i.e. tensor of same size as latent data point filled with 0s and 1s
        t: Optional[Union[torch.nn.ModuleList, torch.nn.Sequential]], optional
            translation mapping, i.e. neural network, where first input dimension is batch dim,
            if None no translation is applied
        s: Optional[Union[torch.nn.ModuleList, torch.nn.Sequential]], optional
            scale mapping, i.e. neural network, where first input dimension is batch dim,
            if None no scale is applied
        """
        super().__init__()
        # self.b = b
        self.b_cpu = b.view(1, *b.size())
        self.register_buffer("b", self.b_cpu)

        if s is None:
            self.s = lambda x: torch.zeros_like(x)
        else:
            self.add_module("s", s)

        if t is None:
            self.t = lambda x: torch.zeros_like(x)
        else:
            self.add_module("t", t)

    def forward(self, z: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """Forward pass of the Masked Affine Flow layer

        Parameters
        ----------
        z : torch.Tensor
            Input tensor

        Returns
        -------
        z : torch.Tensor
            Transformed tensor according to Masked Affine Flow layer with the shape of 'z'.
        log_det : torch.Tensor
            Tensor which represents the information of the deviation of the initial
            and target distribution.
        """
        z_masked: torch.Tensor = self.b * z
        scale = self.s(z_masked)
        nan = torch.tensor(np.nan, dtype=z.dtype, device=z.device)
        scale = torch.where(torch.isfinite(scale), scale, nan)
        trans = self.t(z_masked)
        trans = torch.where(torch.isfinite(trans), trans, nan)
        z_ = z_masked + (1 - self.b) * (z * torch.exp(scale) + trans)
        log_det = torch.sum((1 - self.b) * scale,
                            dim=list(range(1, self.b.dim())))
        return z_, log_det

    def inverse(self, z: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """Inverse pass of the Masked Affine Flow layer

        Parameters
        ----------
        z : torch.Tensor
            Input tensor

        Returns
        -------
        z_ : torch.Tensor
            Transformed tensor according to Masked Affine Flow layer with the shape of 'z'.
        log_det : torch.Tensor
            Tensor which represents the information of the deviation of the initial
            and target distribution.
        """
        z_masked = self.b * z
        scale = self.s(z_masked)
        nan = torch.tensor(np.nan, dtype=z.dtype, device=z.device)
        scale = torch.where(torch.isfinite(scale), scale, nan)
        trans = self.t(z_masked)
        trans = torch.where(torch.isfinite(trans), trans, nan)
        z_ = z_masked + (1 - self.b) * (z - trans) * torch.exp(-scale)
        log_det = -torch.sum(
            (1 - self.b) * scale, dim=list(range(1, self.b.dim())))
        return z_, log_det


class ActNorm(Affine):
    """
    This class implements the ActNorm layer (for activation normalizaton)

    ActNorm is an Affine layer but with a data-dependent initialization,
    where on the very first batch we clever initialize the scale,shift so that the output
    is unit gaussian. As described in [glow]_ Kingma et al (2018).

    ActNorm is a layer that performs an affine transformation of the activations using a scale
    and bias parameter per channel, similar to batch normalization.
    These parameters are initialized such that the post-actnorm activations
    per-channel have zero mean and unit variance given an initial minibatch of data.
    This is a form of data dependent initialization [weight_norm]_.
    After initialization, the scale and bias are treated as regular trainable parameters
    that are independent of the data.

    Examples
    --------
    Importing necessary libraries

    >>> import torch
    >>> import torch.nn as nn
    >>> import torch.nn.functional as F
    >>> from deepchem.models.torch_models.flows import MaskedAffineFlow
    >>> from torch.distributions import MultivariateNormal

    Creating sample data

    >>> dim = 2
    >>> samples = 96
    >>> data = MultivariateNormal(torch.zeros(dim), torch.eye(dim))
    >>> tensor = data.sample(torch.Size((samples, dim)))

    Initializing the ActNorm layer and performing forward and inverse pass

    >>> actnorm = ActNorm(dim)
    >>> _, log_det_jacobian = actnorm.forward(tensor)
    >>> _, inverse_log_det_jacobian = actnorm.inverse(tensor)
    >>> len(inverse_log_det_jacobian)
    96

    References
    ----------
    .. [glow] Kingma, D. P., & Dhariwal, P. (2018).
        Glow: Generative flow with invertible 1x1 convolutions.
        Advances in neural information processing systems, 31.
    .. [weight_norm] Salimans, T., & Kingma, D. P. (2016).
        Weight normalization: A simple reparameterization to
        accelerate training of deep neural networks.
        Advances in neural information processing systems, 29.
    """

    def __init__(self, *args, **kwargs) -> None:
        """Initializes the ActNorm layer
        """
        super().__init__(*args, **kwargs)
        self.data_dep_init_done = torch.tensor(0.0).detach().cpu()

    def forward(self, z: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """Forward pass of the ActNorm layer

        Parameters
        ----------
        z : torch.Tensor
            Input tensor

        Returns
        -------
        z_ : torch.Tensor
            Transformed tensor according to ActNorm layer with the shape of 'z'.
        log_det : torch.Tensor
            Tensor which represents the information of the deviation of the initial
            and target distribution.
        """
        # first batch is used for initialization, c.f. batchnorm
        if not self.data_dep_init_done > 0.0:  # type: ignore
            assert self.scale is not None and self.shift is not None
            s_init = -torch.log(z.std(dim=self.batch_dims, keepdim=True) + 1e-6)
            self.scale.data = s_init.data
            self.shift.data = (-z.mean(dim=self.batch_dims, keepdim=True) *
                               torch.exp(self.scale)).data
            self.data_dep_init_done = torch.tensor(1.0)
        return super().forward(z)

    def inverse(self, z: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """Inverse pass of the ActNorm layer

        Parameters
        ----------
        z : torch.Tensor
            Input tensor

        Returns
        -------
        z_ : torch.Tensor
            Transformed tensor according to ActNorm layer with the shape of 'z'.
        log_det : torch.Tensor
            Tensor which represents the information of the deviation of the initial
            and target distribution.
        """
        # first batch is used for initialization, c.f. batchnorm
        if not self.data_dep_init_done:  # type: ignore
            assert self.scale is not None and self.shift is not None
            s_init = torch.log(z.std(dim=self.batch_dims, keepdim=True) + 1e-6)
            self.scale.data = s_init.data
            self.shift.data = z.mean(dim=self.batch_dims, keepdim=True).data
            self.data_dep_init_done = torch.tensor(1.0)
        return super().inverse(z)


class ClampExp(nn.Module):
    """
    A non Linearity layer that clamps the input tensor by taking the minimum of the
    exponential of the input multiplied by a lambda parameter and 1.

    .. math:: f(x) = min(exp(\lambda * x), 1)

    Example
    -------
    >>> import torch
    >>> from deepchem.models.torch_models.flows import ClampExp
    >>> lambda_param = 1.0
    >>> clamp_exp = ClampExp(lambda_param)
    >>> input = torch.tensor([-1 ,0.5, 0.6, 0.7])
    >>> clamp_exp(input)
    tensor([0.3679, 1.0000, 1.0000, 1.0000])
    """

    def __init__(self, lambda_param: float = 1.0) -> None:
        """
        Initializes the ClampExp layer

        Parameters
        ----------
        lambda_param : float
            Lambda parameter for the ClampExp layer
        """

        self.lambda_param = lambda_param
        super(ClampExp, self).__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the ClampExp layer

        Parameters
        ----------
        x : torch.Tensor
            Input tensor

        Returns
        -------
        torch.Tensor
            Transformed tensor according to ClampExp layer with the shape of 'x'.
        """
        one = torch.tensor(1.0, device=x.device, dtype=x.dtype)
        return torch.min(torch.exp(self.lambda_param * x), one)


class ConstScaleLayer(nn.Module):
    """
    This layer scales the input tensor by a fixed factor

    Example
    -------
    >>> import torch
    >>> from deepchem.models.torch_models.flows import ConstScaleLayer
    >>> scale = 2.0
    >>> const_scale = ConstScaleLayer(scale)
    >>> input = torch.tensor([1, 2, 3])
    >>> const_scale(input)
    tensor([2., 4., 6.])
    """

    def __init__(self, scale: float = 1.0):
        """
        Initializes the ConstScaleLayer

        Parameters
        ----------
        scale : float
            Scaling factor
        """
        super().__init__()
        self.scale = torch.tensor(scale)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        """Forward pass of the ConstScaleLayer

        Parameters
        ----------
        input : torch.Tensor
            Input tensor

        Returns
        -------
        torch.Tensor
            Scaled tensor
        """
        return input * self.scale


class MLP_flow(nn.Module):
    """
    A Multi-Layer Perceptron (MLP) model for normalizing flows that is
    used as a part of a Normalizing Flow model.
    It is a modified version of the MLP model from `deepchem/deepchem/models/torch_models/layers.py`
    to handle multiple layers

    Example
    -------
    >>> import torch
    >>> from deepchem.models.torch_models.flows import MLP_flow
    >>> layers = [2, 4, 4, 2]
    >>> mlp_flow = MLP_flow(layers)
    >>> input = torch.tensor([1., 2.])
    >>> output = mlp_flow(input)
    >>> output.shape
    torch.Size([2])
    """

    def __init__(
        self,
        layers: list,
        leaky: float = 0.0,
        score_scale: Optional[float] = None,
        output_fn=None,
        output_scale: Optional[float] = None,
        init_zeros: bool = False,
        dropout: Optional[float] = None,
    ):
        """
        Initializes the MLP_flow model

        Parameters
        ----------
        layers : list
            List of layer sizes from start to end
        leaky : float, optional default 0.0
            Slope of the leaky part of the ReLU, if 0.0, standard ReLU is used
        score_scale : float, optional
            Factor to apply to the scores, i.e. output before output_fn
        output_fn : str, optional
            Function to be applied to the output, either None, "sigmoid", "relu", "tanh", or "clampexp"
        output_scale : float, optional
            Rescale outputs if output_fn is specified, i.e. scale * output_fn(out / scale)
        init_zeros : bool, optional
            Flag, if true, weights and biases of last layer are initialized with zeros
            (helpful for deep models, see arXiv 1807.03039)
        dropout : float, optional
            If specified, dropout is done before last layer; if None, no dropout is done
        """
        super().__init__()
        net = nn.ModuleList([])
        for k in range(len(layers) - 2):
            net.append(nn.Linear(layers[k], layers[k + 1]))
            net.append(nn.LeakyReLU(leaky))
        if dropout is not None:
            net.append(nn.Dropout(p=dropout))
        net.append(nn.Linear(layers[-2], layers[-1]))
        if init_zeros:
            nn.init.zeros_(net[-1].weight)
            nn.init.zeros_(net[-1].bias)
        if output_fn is not None:
            if score_scale is not None:
                net.append(ConstScaleLayer(score_scale))
            if output_fn == "sigmoid":
                net.append(nn.Sigmoid())
            elif output_fn == "relu":
                net.append(nn.ReLU())
            elif output_fn == "tanh":
                net.append(nn.Tanh())
            elif output_fn == "clampexp":
                net.append(ClampExp())
            else:
                NotImplementedError("This output function is not implemented.")
            if output_scale is not None:
                net.append(ConstScaleLayer(output_scale))
        self.net = nn.Sequential(*net)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of the MLP_flow model

        Parameters
        ----------
        x : torch.Tensor
            Input tensor

        Returns
        -------
        torch.Tensor
            Transformed tensor according to the MLP_flow model with the shape of 'x'
        """
        return self.net(x)



import pytest
import deepchem as dc


@pytest.mark.torch
def testBatchGroverGraph():
    import torch
    from deepchem.utils.grover import BatchGroverGraph
    grover_featurizer = dc.feat.GroverFeaturizer(
        features_generator=dc.feat.CircularFingerprint())

    smiles = ['CC', 'CCC']
    mol_graphs = grover_featurizer.featurize(smiles)
    mol_graph = mol_graphs[0]

    batched_graph = BatchGroverGraph(mol_graphs)
    components = batched_graph.get_components()
    f_atoms, f_bonds, a2b, b2a, b2revb, a2a, a_scope, b_scope, fg_labels = components

    assert batched_graph.additional_features.shape == (
        2, mol_graph.additional_features.shape[0])
    # 6 atoms: CC -> 2, CCC -> 3
    assert f_atoms.shape == (5, mol_graph.node_features.shape[1])
    # 7 bonds: CC -> 2, CCC -> 4 (bonds are considered as undirected
    # and a single bond contributes to 2 bonds)
    assert f_bonds.shape == (6, mol_graph.edge_features.shape[1])
    assert fg_labels.shape == (2, mol_graph.fg_labels.shape[0])
    assert (a_scope == torch.Tensor([[0, 2], [2, 3]])).all()
    assert (b_scope == torch.Tensor([[0, 2], [2, 4]])).all()

    assert (a2b == torch.Tensor([[0, 0], [1, 0], [0, 0], [3, 0], [5, 0]])).all()
    assert (b2a == torch.Tensor([0, 1, 2, 4, 3, 4])).all()

    assert (b2revb == torch.Tensor([1, 0, 3, 2, 5, 4])).all()
    assert (a2a == torch.Tensor([[0, 0], [1, 0], [0, 0], [4, 0], [4, 0]])).all()



"""
Generate coulomb matrices for molecules.

See Montavon et al., _New Journal of Physics_ __15__ (2013) 095003.

"""
import numpy as np
from typing import Any, List, Optional

from deepchem.utils.typing import RDKitMol
from deepchem.utils.data_utils import pad_array
from deepchem.feat.base_classes import MolecularFeaturizer


class CoulombMatrix(MolecularFeaturizer):
    """Calculate Coulomb matrices for molecules.

    Coulomb matrices provide a representation of the electronic structure of
    a molecule. For a molecule with `N` atoms, the Coulomb matrix is a
    `N X N` matrix where each element gives the strength of the
    electrostatic interaction between two atoms. The method is described
    in more detail in [1]_.

    Examples
    --------
    >>> import deepchem as dc
    >>> featurizers = dc.feat.CoulombMatrix(max_atoms=23)
    >>> input_file = 'deepchem/feat/tests/data/water.sdf' # really backed by water.sdf.csv
    >>> tasks = ["atomization_energy"]
    >>> loader = dc.data.SDFLoader(tasks, featurizer=featurizers)
    >>> dataset = loader.create_dataset(input_file)

    References
    ----------
    .. [1] Montavon, Grégoire, et al. "Learning invariant representations of
        molecules for atomization energy prediction." Advances in neural information
        processing systems. 2012.

    Note
    ----
    This class requires RDKit to be installed.

    """

    def __init__(self,
                 max_atoms: int,
                 remove_hydrogens: bool = False,
                 randomize: bool = False,
                 upper_tri: bool = False,
                 n_samples: int = 1,
                 seed: Optional[int] = None):
        """Initialize this featurizer.

        Parameters
        ----------
        max_atoms: int
            The maximum number of atoms expected for molecules this featurizer will
            process.
        remove_hydrogens: bool, optional (default False)
            If True, remove hydrogens before processing them.
        randomize: bool, optional (default False)
            If True, use method `randomize_coulomb_matrices` to randomize Coulomb matrices.
        upper_tri: bool, optional (default False)
            Generate only upper triangle part of Coulomb matrices.
        n_samples: int, optional (default 1)
            If `randomize` is set to True, the number of random samples to draw.
        seed: int, optional (default None)
            Random seed to use.

        """
        self.max_atoms = int(max_atoms)
        self.remove_hydrogens = remove_hydrogens
        self.randomize = randomize
        self.upper_tri = upper_tri
        self.n_samples = n_samples
        if seed is not None:
            seed = int(seed)
        self.seed = seed

    def _featurize(self, datapoint: RDKitMol, **kwargs) -> np.ndarray:
        """
        Calculate Coulomb matrices for molecules. If extra randomized
        matrices are generated, they are treated as if they are features
        for additional conformers.

        Since Coulomb matrices are symmetric, only the (flattened) upper
        triangular portion is returned.

        Parameters
        ----------
        datapoint: rdkit.Chem.rdchem.Mol
            RDKit Mol object

        Returns
        -------
        np.ndarray
            The coulomb matrices of the given molecule.
            The default shape is `(num_confs, max_atoms, max_atoms)`.
            If num_confs == 1, the shape is `(max_atoms, max_atoms)`.

        """
        if 'mol' in kwargs:
            datapoint = kwargs.get("mol")
            raise DeprecationWarning(
                'Mol is being phased out as a parameter, please pass "datapoint" instead.'
            )

        features = self.coulomb_matrix(datapoint)
        if self.upper_tri:
            features = np.asarray(
                [f[np.triu_indices_from(f)] for f in features])
        if features.shape[0] == 1:
            # `(1, max_atoms, max_atoms)` -> `(max_atoms, max_atoms)`
            features = np.squeeze(features, axis=0)
        return features

    def coulomb_matrix(self, mol: RDKitMol) -> np.ndarray:
        """
        Generate Coulomb matrices for each conformer of the given molecule.

        Parameters
        ----------
        mol: rdkit.Chem.rdchem.Mol
            RDKit Mol object

        Returns
        -------
        np.ndarray
            The coulomb matrices of the given molecule

        """
        try:
            from rdkit import Chem
            from rdkit.Chem import AllChem
        except ModuleNotFoundError:
            raise ImportError("This class requires RDKit to be installed.")

        # Check whether num_confs >=1 or not
        num_confs = len(mol.GetConformers())
        if num_confs == 0:
            mol = Chem.AddHs(mol)
            AllChem.EmbedMolecule(mol, AllChem.ETKDG())

        if self.remove_hydrogens:
            mol = Chem.RemoveHs(mol)
        n_atoms = mol.GetNumAtoms()
        z = [atom.GetAtomicNum() for atom in mol.GetAtoms()]
        rval = []
        for conf in mol.GetConformers():
            d = self.get_interatomic_distances(conf)
            m = np.outer(z, z) / d
            m[range(n_atoms), range(n_atoms)] = 0.5 * np.array(z)**2.4
            if self.randomize:
                for random_m in self.randomize_coulomb_matrix(m):
                    random_m = pad_array(random_m, self.max_atoms)
                    rval.append(random_m)
            else:
                m = pad_array(m, self.max_atoms)
                rval.append(m)
        return np.asarray(rval)

    def randomize_coulomb_matrix(self, m: np.ndarray) -> List[np.ndarray]:
        """Randomize a Coulomb matrix as decribed in [1]_:

        1. Compute row norms for M in a vector row_norms.
        2. Sample a zero-mean unit-variance noise vector e with dimension
            equal to row_norms.
        3. Permute the rows and columns of M with the permutation that
            sorts row_norms + e.

        Parameters
        ----------
        m: np.ndarray
            Coulomb matrix.

        Returns
        -------
        List[np.ndarray]
            List of the random coulomb matrix

        References
        ----------
        .. [1] Montavon et al., New Journal of Physics, 15, (2013), 095003

        """
        rval = []
        row_norms = np.asarray([np.linalg.norm(row) for row in m], dtype=float)
        rng = np.random.RandomState(self.seed)
        for i in range(self.n_samples):
            e = rng.normal(size=row_norms.size)
            p = np.argsort(row_norms + e)
            new = m[p][:, p]  # permute rows first, then columns
            rval.append(new)
        return rval

    @staticmethod
    def get_interatomic_distances(conf: Any) -> np.ndarray:
        """
        Get interatomic distances for atoms in a molecular conformer.

        Parameters
        ----------
        conf: rdkit.Chem.rdchem.Conformer
            Molecule conformer.

        Returns
        -------
        np.ndarray
            The distances matrix for all atoms in a molecule

        """
        n_atoms = conf.GetNumAtoms()
        coords = [
            # Convert AtomPositions from Angstrom to bohr (atomic units)
            conf.GetAtomPosition(i).__idiv__(0.52917721092)
            for i in range(n_atoms)
        ]
        d = np.zeros((n_atoms, n_atoms), dtype=float)
        for i in range(n_atoms):
            for j in range(i):
                d[i, j] = coords[i].Distance(coords[j])
                d[j, i] = d[i, j]
        return d


class CoulombMatrixEig(CoulombMatrix):
    """Calculate the eigenvalues of Coulomb matrices for molecules.

    This featurizer computes the eigenvalues of the Coulomb matrices for provided
    molecules. Coulomb matrices are described in [1]_.

    Examples
    --------
    >>> import deepchem as dc
    >>> featurizers = dc.feat.CoulombMatrixEig(max_atoms=23)
    >>> input_file = 'deepchem/feat/tests/data/water.sdf' # really backed by water.sdf.csv
    >>> tasks = ["atomization_energy"]
    >>> loader = dc.data.SDFLoader(tasks, featurizer=featurizers)
    >>> dataset = loader.create_dataset(input_file)

    References
    ----------
    .. [1] Montavon, Grégoire, et al. "Learning invariant representations of
        molecules for atomization energy prediction." Advances in neural information
        processing systems. 2012.

    """

    def __init__(self,
                 max_atoms: int,
                 remove_hydrogens: bool = False,
                 randomize: bool = False,
                 n_samples: int = 1,
                 seed: Optional[int] = None):
        """Initialize this featurizer.

        Parameters
        ----------
        max_atoms: int
            The maximum number of atoms expected for molecules this featurizer will
            process.
        remove_hydrogens: bool, optional (default False)
            If True, remove hydrogens before processing them.
        randomize: bool, optional (default False)
            If True, use method `randomize_coulomb_matrices` to randomize Coulomb matrices.
        n_samples: int, optional (default 1)
            If `randomize` is set to True, the number of random samples to draw.
        seed: int, optional (default None)
            Random seed to use.

        """
        self.max_atoms = int(max_atoms)
        self.remove_hydrogens = remove_hydrogens
        self.randomize = randomize
        self.n_samples = n_samples
        if seed is not None:
            seed = int(seed)
        self.seed = seed

    def _featurize(self, datapoint: RDKitMol, **kwargs) -> np.ndarray:
        """
        Calculate eigenvalues of Coulomb matrix for molecules. Eigenvalues
        are returned sorted by absolute value in descending order and padded
        by max_atoms.

        Parameters
        ----------
        datapoint: rdkit.Chem.rdchem.Mol
            RDKit Mol object

        Returns
        -------
        np.ndarray
            The eigenvalues of Coulomb matrix for molecules.
            The default shape is `(num_confs, max_atoms)`.
            If num_confs == 1, the shape is `(max_atoms,)`.

        """
        if 'mol' in kwargs:
            datapoint = kwargs.get("mol")
            raise DeprecationWarning(
                'Mol is being phased out as a parameter, please pass "datapoint" instead.'
            )

        cmat = self.coulomb_matrix(datapoint)
        features_list = []
        for f in cmat:
            w, v = np.linalg.eig(f)
            w_abs = np.abs(w)
            sortidx = np.argsort(w_abs)
            sortidx = sortidx[::-1]
            w = w[sortidx]
            f = pad_array(w, self.max_atoms)
            features_list.append(f)
        features = np.asarray(features_list)
        if features.shape[0] == 1:
            # `(1, max_atoms)` -> `(max_atoms,)`
            features = np.squeeze(features, axis=0)
        return features



"""
Featurizers for inorganic crystals.
"""
# flake8: noqa
from deepchem.feat.material_featurizers.element_property_fingerprint import ElementPropertyFingerprint
from deepchem.feat.material_featurizers.sine_coulomb_matrix import SineCoulombMatrix
from deepchem.feat.material_featurizers.cgcnn_featurizer import CGCNNFeaturizer
from deepchem.feat.material_featurizers.elemnet_featurizer import ElemNetFeaturizer
from deepchem.feat.material_featurizers.lcnn_featurizer import LCNNFeaturizer



from __future__ import annotations
from abc import abstractmethod, abstractproperty
from typing import List
import torch
from deepchem.utils.differentiation_utils import EditableModule, LinearOperator


class BaseDF(EditableModule):
    """
    BaseDF represents the density fitting object used in calculating the
    electron repulsion (and xc energy) in Hamiltonian.

    Density fitting in density functional theory (DFT) is a technique used to
    reduce the computational cost of evaluating electron repulsion integrals.
    In DFT, the key quantity is the electron density rather than the wave
    function, and the electron repulsion integrals involve four-electron
    interactions, making them computationally demanding.

    Density fitting exploits the fact that many-electron integrals can be
    expressed as a sum of two-electron integrals involving auxiliary basis
    functions. By approximating these auxiliary basis functions, often referred
    to as fitting functions, the computational cost can be significantly reduced.

    Examples
    --------
    >>> from deepchem.utils.dft_utils import BaseDF
    >>> import torch
    >>> class MyDF(BaseDF):
    ...     def __init__(self):
    ...         super(MyDF, self).__init__()
    ...     def get_j2c(self):
    ...         return torch.ones((3, 3))
    ...     def get_j3c(self):
    ...         return torch.ones((3, 3, 3))
    >>> df = MyDF()
    >>> df.get_j2c()
    tensor([[1., 1., 1.],
            [1., 1., 1.],
            [1., 1., 1.]])

    """

    @abstractmethod
    def build(self) -> BaseDF:
        """
        Construct the matrices required to perform the calculation and return
        self.

        Returns
        -------
        BaseDF
            The constructed density fitting object.

        """
        pass

    @abstractmethod
    def get_elrep(self, dm: torch.Tensor) -> LinearOperator:
        """
        Construct the electron repulsion linear operator from the given density
        matrix using the density fitting method.

        Parameters
        ----------
        dm : torch.Tensor
            The density matrix.

        Returns
        -------
        LinearOperator
            The electron repulsion linear operator.

        """
        pass

    # properties
    @abstractproperty
    def j2c(self) -> torch.Tensor:
        """Returns the 2-centre 2-electron integrals of the auxiliary basis.

        Returns
        -------
        torch.Tensor
            The 2-centre 2-electron integrals of the auxiliary basis.

        """
        pass

    @abstractproperty
    def j3c(self) -> torch.Tensor:
        """
        Return the 3-centre 2-electron integrals of the auxiliary basis and the
        basis.

        Returns
        -------
        torch.Tensor
            The 3-centre 2-electron integrals of the auxiliary basis and the
            basis.

        """
        pass

    @abstractmethod
    def getparamnames(self, methodname: str, prefix: str = "") -> List[str]:
        """
        This method should list tensor names that affect the output of the
        method with name indicated in ``methodname``.

        Parameters
        ---------
        methodname: str
            The name of the method of the class.
        prefix: str (default="")
            The prefix to be appended in front of the parameters name.
            This usually contains the dots.

        Returns
        -------
        List[str]
            Sequence of name of parameters affecting the output of the method.

        """
        pass



"""Interface for reinforcement learning."""
# flake8: noqa
try:
    from deepchem.rl.torch_rl.torch_a2c import A2CLossDiscrete, A2CLossContinuous, A2C  # noqa: F401
    from deepchem.rl.torch_rl.torch_ppo import PPOLoss, PPO  # noqa: F401
except ModuleNotFoundError:
    pass
